from pydantic import BaseModel, StrictStr
from .msk_cluster import MSKCluster


class MSKListClustersResponse(BaseModel):
    ClusterInfoList: list[MSKCluster]
    NextToken: StrictStr | None = None
    


from pydantic import BaseModel, StrictStr, model_validator
from typing import Literal, Self

GroupTypes = Literal[
    "DIMENSION",
    "TAG",
    "COST_CATEGORY"
]

DimensionTypes = Literal[
    "AZ",
    "INSTANCE_TYPE",
    "LEGAL_ENTITY_NAME",
    "INVOICING_ENTITY",
    "LINKED_ACCOUNT",
    "OPERATION",
    "PLATFORM",
    "PURCHASE_TYPE",
    "SERVICE",
    "TENANCY",
    "RECORD_TYPE",
    "USAGE_TYPE",
]

class Group(BaseModel):
    Type: GroupTypes
    Key: DimensionTypes | StrictStr

    @model_validator(mode='after')
    def validate_dimension_types(self) -> Self:
        dimension_types = [
            "AZ",
            "INSTANCE_TYPE",
            "LEGAL_ENTITY_NAME",
            "INVOICING_ENTITY",
            "LINKED_ACCOUNT",
            "OPERATION",
            "PLATFORM",
            "PURCHASE_TYPE",
            "SERVICE",
            "TENANCY",
            "RECORD_TYPE",
            "USAGE_TYPE",
        ]

        if self.Type == 'DIMENSION' and self.Key not in dimension_types:
            valid_types = ', '.join(dimension_types)
            raise ValueError(f'Invalid Dimension type {self.Key} - must be one of {valid_types}')
        
        return self
    
    def dump(self):
        return self.model_dump()


from pydantic import BaseModel, StrictStr
from .cloudwatch_datapoint import CloudWatchDatapoint


class CloudWatchGetMetricStatisticsResponse(BaseModel):
    Label: StrictStr
    Datapoints: list[CloudWatchDatapoint]
    


from typing import Literal
from pydantic import BaseModel, StrictStr, StrictBool
from .cloudwatch_dimension import CloudWatchDimension


class CloudWatchListMetricsRequest(BaseModel):
    Namespace: StrictStr | None = None
    MetricName: StrictStr | None = None
    Dimensions: list[CloudWatchDimension] | None = None
    NextToken: StrictStr | None = None
    RecentlyActive: Literal["PT3H"] | None = None
    IncludeLinkedAccounts: StrictBool = False
    OwningAccount: StrictStr | None = None


from .elasticache_describe_cache_clusters_request import ElasticacheDescribeCacheClustersRequest as ElasticacheDescribeCacheClustersRequest
from .elasticache_describe_cache_clusters_response import ElasticacheDescribeCacheClustersResponse as ElasticacheDescribeCacheClustersResponse
from .elasticache_describe_replication_groups_request import ElasticacheDescribeReplicationGroupsRequest as ElasticacheDescribeReplicationGroupsRequest
from .elasticache_describe_replication_groups_response import ElasticacheDescribeReplicationGroupsResponse as ElasticacheDescribeReplicationGroupsResponse


from pydantic import BaseModel, StrictBool
from .msk_client_broker import MSKClientBroker


class MSKEncryptionInTransit(BaseModel):
    ClientBroker: MSKClientBroker
    InCluster: StrictBool = False


from pydantic import BaseModel
from .msk_client_authentication_sasl_iam import MSKClientAuthenticationSaslIAm
from .msk_client_authentication_sasl_scram import MSKClientAuthenticationSaslScram

class MSKClientAuthenticationSasl(BaseModel):
    Scram: MSKClientAuthenticationSaslScram
    Iam: MSKClientAuthenticationSaslIAm


from .ec2_describe_volumes_query import EC2DescribeVolumesQuery as EC2DescribeVolumesQuery
from .ec2_describe_volumes_response import EC2DescribeVolumesResponse as EC2DescribeVolumesResponse
from .ec2_filter import EC2Filter as EC2Filter
from .ec2_operations import EC2Operations as EC2Operations


from abc import ABC, abstractmethod
from typing import Dict


class CodeArtifactClient(ABC):

    @abstractmethod
    def get_authorization_token(
        self,
        domain: str=None,
        domainOwner: str=None,
        durationSeconds: int=None
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def get_repository_endpoint(
        self,
        domain: str=None,
        domainOwner: str=None,
        repository: str=None,
        format: str=None
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def close(self):
        pass


from enum import Enum
from typing import Literal


CloudWatchServiceName = Literal[
    'access-analyzer',
    'amazon-emr',
    'amazon-lex',
    'amazon-personalize',
    'amazon-s3-glacier',
    'amazon-workmail',
    'api-gateway',
    'appflow',
    'appintegrations',
    'appstream',
    'athena',
    'audit-manager',
    'aws-appconfig',
    'aws-codeartifact',
    'aws-config',
    'aws-health-imaging',
    'aws-user-notifications',
    'backup',
    'batch',
    'bedrock',
    'certificate-manager',
    'clean-rooms',
    'cloud-map',
    'cloudformation',
    'cloudhsm',
    'cloudshell',
    'cloudtrail',
    'cloudwatch',
    'cloudwatch-application-insights',
    'codebuild',
    'codecommit',
    'code-pipeline',
    'codestart-notifications',
    'cognito-user-pool',
    'comprehend',
    'computeoptimizer',
    'connect',
    'customer-profiles',
    'data-exchange',
    'data-lifecycle-manager',
    'database-migration-service',
    'databrew',
    'datasync',
    'directive',
    'directory-service',
    'dynamodb',
    'dynamodb-accelerator',
    'ec2',
    'ec2-auto-scaling',
    'ecr',
    'ecs',
    'efs',
    'elastic-beanstalk',
    'elastic-load-balancing',
    'elasticache',
    'emr',
    'emr-containers',
    'emr-serverless',
    'eventbridge',
    'fraud-detector',
    'gramelift',
    'grafana',
    'greengrass',
    'ground-station',
    'guardduty',
    'image-builder',
    'iot',
    'iot-events',
    'iot-sitwise',
    'iot-twinmaker',
    'iot-wireless',
    'ivs',
    'kendra',
    'kinesis-analytics',
    'kinesis-data-streams',
    'kinesis-video-streams',
    'kms',
    'lambda',
    'lightsail',
    'location',
    'logs',
    'macie',
    'mediaconnect',
    'mediaconvert',
    'memorydb',
    'neptunegraph',
    'network-firewall',
    'outposts',
    'pinpoint',
    'private-certificate-authority',
    'qldb',
    'rds',
    'redshift',
    'resource-access-manager',
    'resource-groups',
    'resource-groups-tagging-api',
    'robo-maker',
    'route-53-resolver',
    's3',
    'sagemaker',
    'scheduler',
    'secrets-manager',
    'security-hub',
    'service-catalog-appregistry',
    'ses',
    'sns',
    'ssm',
    'storage-gateway',
    'transcribe',
    'transfer',
    'translate',
    'workspaces',
    'x-ray'
]


class CloudWatchAWSService(Enum):
    ACCESS_ANALYZER='Access Analyzer'
    AMAZON_EMR='Amazon EMR'
    AMAZON_LEX='Amazon Lex'
    AMAZON_PERSONALIZE='Amazon Personalize'
    AMAZON_S3_GLACIER='Amazon S3 Glacier'
    AMAZON_WORKMAIL='Amazon WorkMail'
    API_GATEWAY='API Gateway'
    APPFLOW='AppFlow'
    APPINTEGRATIONS='AppIntegrations'
    APPSTREAM='AppStream'
    ATHENA='Athena'
    AUDIT_MANAGER='Audit Manager'
    AWS_APPCONFIG='AWS AppConfig'
    AWS_CODEARTIFACT='AWS CodeArtifact'
    AWS_CONFIG='AWS Config'
    AWS_HEALTH_IMAGING='AWS Health Imaging'
    AWS_USER_NOTIFICATIONS='AWS User Notifications'
    BACKUP='Backup'
    BATCH='Batch'
    BEDROCK='Bedrock'
    CERTIFICATE_MANGER='Certificate Manager'
    CLEAN_ROOMS='Clean Rooms'
    CLOUD_MAP='Cloud Map'
    CLOUDFORMATION='CloudFormation'
    CLOUDHSM='CloudHSM'
    CLOUDSHELL='CloudShell'
    CLOUDTRAIL='CloudTrail'
    CLOUDWATCH='CloudWatch'
    CLOUDWATCH_APPLICATION_INSIGHTS='CloudWatch Application Insights'
    CODEBUILD='CodeBuild'
    CODECOMMIT='CodeCommit'
    CODE_PIPELINE='CodePipeline'
    CODESTART_NOTIFICATIONS='CodeStar Notifications'
    COGNITO_USER_POOL='Cognito User Pool'
    COMPREHEND='Comprehend'
    COMPUTEOPTIMIZER='ComputeOptimizer'
    CONNECT='Connect'
    CUSTOMER_PROFILES='Customer Profiles'
    DATA_EXCHANGE='Data Exchange'
    DATA_LIFECYCLE_MANAGER='Data Lifecycle Manager'
    DATABASE_MIGRATION_SERVICE='Database Migration Service'
    DATABREW='DataBrew'
    DATASYNC='DataSync'
    DIRECTIVE='Directive'
    DIRECTORY_SERVICE='Directory Service'
    DYNAMODB='DynamoDB'
    DYNAMODB_ACCELERATOR='DynamoDB Accelerator'
    EC2='EC2'
    EC2_AUTO_SCALING='EC2 Auto Scaling'
    ECR='ECR'
    ECS='ECS'
    EFS='EFS'
    ELASTIC_BEANSTALK='Elastic Beanstalk'
    ELASTIC_LOAD_BALANCING='Elastic Load Balancing'
    ELASTICACHE='ElastiCache'
    EMR='EMR'
    EMR_CONTAINERS='EMR Containers'
    EMR_SERVERLESS='EMR Serverless'
    EVENTBRIDGE='EventBridge'
    FRAUD_DETECTOR='Fraud Detector'
    GAMELIFT='GameLift'
    GRAFANA='Grafana'
    GREENGRASS='Greengrass'
    GROUND_STATION='Ground Station'
    GUARDDUTY='GuardDuty'
    IMAGE_BUILDER='Image Builder'
    IOT='IoT'
    IOT_EVENTS='IoT Events'
    IOT_SITEWISE='IoT SiteWise'
    IOT_TWINMAKER='IoT TwinMaker'
    IOT_WIRELESS='IoT Wireless'
    IVS='IVS'
    KENDRA='Kendra'
    KINESIS_ANALYTICS='Kinesis Analytics'
    KINESIS_DATA_STREAMS='Kinesis Data Streams'
    KINESIS_VIDEO_STREAMS='Kinesis Video Streams'
    KMS='KMS'
    LAMBDA='Lambda'
    LIGHTSAIL='Lightsail'
    LOCATION='Location'
    LOGS='Logs'
    MACIE='Macie'
    MEDIACONNECT='MediaConnect'
    MEDIACONVERT='MediaConvert'
    MEMORYDB='MemoryDB'
    NEPTUNEGRAPH='NeptuneGraph'
    NETWORK_FIREWALL='Network Firewall'
    OUTPOSTS='Outposts'
    PINPOINT='Pinpoint'
    PRIVATE_CERTIFICATE_AUTHORITY='Private Certificate Authority'
    QLDB='QLDB'
    RDS='RDS'
    REDSHIFT='Redshift'
    RECOURCE_ACCESS_MANAGER='Resource Access Manager'
    RESOURCE_GROUPS='Resource Groups'
    RESOURCE_GROUPS_TAGGING_API='Resource Groups Tagging API'
    ROBO_MAKER='RoboMaker'
    ROUTE_53_RESOLVER='Route 53 Resolver'
    S3='S3'
    SAGEMAKER='SageMaker'
    SCHEDULER='Scheduler'
    SECRETS_MANAGER='Secrets Manager'
    SECURITY_HUB='Security Hub'
    SERVICE_CATALOG_APPREGISTRY='Service Catalog AppRegistry'
    SES='SES'
    SNS='SNS'
    SSM='SSM'
    STORAGE_GATEWAY='Storage Gateway'
    TRANSCRIBE='Transcribe'
    TRANSFER='Transfer'
    TRANSLATE='Translate'
    WORKSPACES='WorkSpaces'
    X_RAY='X-Ray'



class CloudWatchAWSServicesMap:

    def __init__(self):
        self._services_map: dict[
            CloudWatchServiceName,
            CloudWatchAWSService
        ] = {
            service.name.lower().replace(
                '_',
                '-'
            ): service for service in CloudWatchAWSService
        }

    def get(self, service_name: CloudWatchServiceName):
        return self._services_map[service_name]


from pydantic import (
    BaseModel,
    StrictInt, 
    StrictBool
)
from typing import Literal, Optional


class AWSs3TransferConfigOptions(BaseModel):
    multipart_threshold: Optional[StrictInt]=None
    max_concurrency: Optional[StrictInt]=None
    multipart_chunksize: Optional[StrictInt]=None
    num_download_attmpets: Optional[StrictInt]=None
    max_io_queue: Optional[StrictInt]=None
    io_chunksize: Optional[StrictInt]=None
    use_threads: Optional[StrictBool]=False
    max_bandwidth: Optional[StrictInt]=None
    preferred_trasfer_client: Optional[
        Literal[
            'auto',
            'classic'
        ]
    ]='auto'


from pydantic import BaseModel, StrictStr


class AWSConfigRole(BaseModel):
    role_name: StrictStr
    region: StrictStr
    source_profile: StrictStr
    external_id: StrictStr
    role_arn: StrictStr


from pydantic import BaseModel
from .msk_jmx_exporter import MSKJmxExporter
from .msk_node_exporter import MSKNodeExporter


class MSKPrometheus(BaseModel):
    JmxExporter: MSKJmxExporter
    NodeExporter: MSKNodeExporter



class EmptyResponseException(Exception):

    def __init__(
        self,
        service_name: str,
        call_name: str,
        retrieval_item_name: str
    ) -> None:
        super().__init__(
            f'Err. - call to - {call_name} - for service - {service_name} - failed to retrieve - {retrieval_item_name} - and returned None.'
        )


from pydantic import (
    BaseModel,
    StrictStr,
    StrictBool,
)


class MSKS3(BaseModel):
    Bucket: StrictStr | None = None
    Enabled: StrictBool = False
    Prefix: StrictStr | None = None
    


from pydantic import BaseModel, StrictStr
from .elasticache_global_replication_group_info import ElasticacheGlobalReplicationGroupInfo
from .elasticache_pending_modified_values import ElasticachePendingModifiedValues

class ElasticacheReplicationGroup(BaseModel):
    ReplicationGroupId: StrictStr
    Description: StrictStr
    GlobalReplicationGroupInfo: ElasticacheGlobalReplicationGroupInfo
    Status: StrictStr
    PendingModifiedValues: ElasticachePendingModifiedValues | None = None
    MemberClusters: list[StrictStr] | None = None


from pydantic import BaseModel
from .msk_connectivity_public_access import MSKConnectivityPublicAccess
from .msk_vpc_connectivity import MSKVpcConnectivity


class MSKConnectivityInfo(BaseModel):
    PublicAccess: MSKConnectivityPublicAccess
    VpcConnectivity: MSKVpcConnectivity



from .cloudwatch_api_operations_map import CloudWatchAPIOperationName as CloudWatchAPIOperationName
from .cloudwatch_api_operations_map import CloudServiceAPIOpsMap as CloudServiceAPIOpsMap
from .cloudwatch_api_operations_map import CloudWatchAPIOperation as CloudWatchAPIOperation
from .cloudwatch_aws_services_map import CloudWatchServiceName as CloudWatchServiceName
from .cloudwatch_aws_services_map import CloudWatchAWSService as CloudWatchAWSService
from .cloudwatch_aws_services_map import CloudWatchAWSServicesMap as CloudWatchAWSServicesMap
from .cloudwatch_dimension import CloudWatchDimension as CloudWatchDimension
from .cloudwatch_list_metrics_request import CloudWatchListMetricsRequest as CloudWatchListMetricsRequest
from .cloudwatch_list_metrics_response import CloudWatchListMetricsResponse as CloudWatchListMetricsResponse
from .cloudwatch_get_metric_statistics_request import CloudWatchGetMetricStatisticsRequest as CloudWatchGetMetricStatisticsRequest
from .cloudwatch_get_metric_statistics_response import CloudWatchGetMetricStatisticsResponse as CloudWatchGetMetricStatisticsResponse
from .cloudwatch_get_metric_data_request import CloudWatchGetMetricDataRequest as CloudWatchGetMetricDataRequest
from .cloudwatch_metric_data_query import CloudWatchMetricDataQuery as CloudWatchMetricDataQuery
from .cloudwatch_metric_stat import CloudWatchMetricStat as CloudWatchMetricStat
from .cloudwatch_metric import CloudWatchMetric as CloudWatchMetric
from .cloudwatch_get_metric_data_response import CloudWatchGetMetricDataResponse as CloudWatchGetMetricDataResponse
from .cloudwatch_datapoint import CloudWatchDatapoint as CloudWatchDatapoint
from .cloudwatch_metric_data_result import CloudWatchMetricDataResult as CloudWatchMetricDataResult


from pydantic import (
    BaseModel,
    StrictStr,
    StrictBool
)

from typing import (
    List,
    Dict,
    Literal,
    Optional
)
from .aws_s3_deleted_object import AWSs3DeletedObject
from .aws_s3_deleted_object_error import AWSs3DeleteObjectError


class AWSs3DeleteObjectsResponse(BaseModel):
    Deleted: List[AWSs3DeletedObject]=[]
    RequestCharged: Optional[
        Literal['requester']
    ]=None
    Errors: List[AWSs3DeleteObjectError]=[]



from .aws_ec2 import AWSEC2 as AWSEC2
from .aws_cloudwatch import AWSCloudwatch as AWSCloudwatch
from .aws_codeartifact import AWSCodeArtifact as AWSCodeArtifact
from .aws_credentials import AWSCredentials as AWSCredentials
from .aws_cost_explorer import AWSCostExplorer as AWSCostExplorer
from .aws_elasticache import AWSElastiCache as AWSElastiCache
from .aws_msk import AWSMSK as AWSMSK
from .aws_organizations import AWSOrganizations as AWSOrganizations
from .aws_sts import AWSSTS as AWSSTS


import datetime
from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictBool
)
from typing import Literal, Dict, List


class AWSs3ObjectResponse(BaseModel):
    Key: StrictStr
    LastModified: datetime.datetime
    ETag: StrictStr
    ChecksumAlgorithm: List[
        Literal[
            'CRC32',
            'CRC32C',
            'SHA1',
            'SHA256'
        ]
    ]=[]
    StrictInt: StrictInt
    StorageClass: Literal[
        'STANDARD',
        'REDUCED_REDUNDANCY',
        'GLACIER',
        'STANDARD_IA',
        'ONEZONE_IA',
        'INTELLIGENT_TIERING',
        'DEEP_ARCHIVE',
        'OUTPOSTS',
        'GLACIER_IR',
        'SNOW',
        'EXPRESS_ONEZONE'
    ]
    Owner: Dict[
        Literal['DisplayName', 'ID'],
        StrictStr
    ]
    RestoreStatus: Dict[
        Literal[       
            'IsRestoreInProgress',
            'RestoreExpiryDate'
        ],
        StrictBool, datetime.datetime
    ]


from __future__ import annotations
from abc import ABC, abstractmethod
from botocore.paginate import Paginator
from typing import Any

class ElastiCacheClient(ABC):

    @abstractmethod
    def describe_cache_clusters(
        self,
        CacheClusterId: str = None,
        MaxRecords: int = None,
        Marker: str = None,
        ShowCacheNodeInfo: bool = True,
        ShowCacheClustersNotInReplicationGroups: bool = True,
    ) -> dict[str, Any]:
        pass
    
    @abstractmethod
    def describe_replication_groups(
        self,
        ReplicationGroupId: str = None,
        MaxRecords: int = None,
        Marker: str = None,
    ) -> dict[str, Any]:
        pass

    @abstractmethod
    def get_paginator(
        self,
        Operation: str,
    ) -> Paginator:
        pass

    @abstractmethod
    def close(self):
        pass


from pydantic import (
    BaseModel,
    StrictStr
)

from .organizations_account import OrganizationsAccount


class OrganizationsListAccountsResponse(BaseModel):
    Accounts: list[OrganizationsAccount]
    NextToken: StrictStr | None = None


from pydantic import BaseModel
from .elasticache_cloud_watch_logs_detail import ElasticacheCloudWatchLogsDetail
from .elasticache_kinesis_firehose_delivery_detail import ElasticacheKinesisFirehoseDetail


class ElasticacheDestinationDetails(BaseModel):
    CloudWatchLogsDetails: ElasticacheCloudWatchLogsDetail | None = None
    KinesisFirehoseDetails: ElasticacheKinesisFirehoseDetail | None = None


from pydantic import BaseModel
from .msk_client_authentication_sasl import MSKClientAuthenticationSasl
from .msk_vpc_client_authentication_tls import MSKVPCClientAuthenticationTls


class MSKVPCClientAuthentication(BaseModel):
    Sasl: MSKClientAuthenticationSasl
    Tls: MSKVPCClientAuthenticationTls




from pydantic import BaseModel, StrictInt
from .msk_ebs_provisioned_throughtput import MSKEBSPRovisionedThroughtput


class MSKEBSStorageInfo(BaseModel):
    ProvisionedThroughput: MSKEBSPRovisionedThroughtput
    VolumeSize: StrictInt


import datetime
from pydantic import (
    BaseModel,
    StrictStr,
)
from .organizations_joined_method import AWSJoinedMethod
from .organizations_account_status import AWSAccountStatus

class OrganizationsAccount(BaseModel):
    Id: StrictStr
    Arn: StrictStr
    Email: StrictStr
    Name: StrictStr
    Status: AWSAccountStatus
    JoinedMethod: AWSJoinedMethod
    JoinedTimestamp: datetime.datetime



import datetime
from pydantic import (
    StrictStr,
    StrictInt
)

from typing import Optional, Literal
from werkflow_aws.models.base import AWSBoto3Base
from werkflow_aws.models.parsing import (
    convert_key_to_boto3_arg, 
    convert_key_to_boto3_arg_upper_matching
)



class AWSs3GetObjectOptions(AWSBoto3Base):
    if_match: Optional[StrictStr]=None
    if_modified_since: Optional[datetime.datetime]=None
    if_none_match: Optional[StrictStr]=None
    if_unmodified_since: Optional[datetime.datetime]=None
    range: Optional[StrictStr]=None
    response_cache_control: Optional[StrictStr]=None
    response_content_disposition: Optional[StrictStr]=None
    response_content_encoding: Optional[StrictStr]=None
    response_content_language: Optional[StrictStr]=None
    response_content_type: Optional[StrictStr]=None
    response_expires: Optional[datetime.datetime]=None
    version_id: Optional[StrictStr]=None
    sse_customer_algorithm: Optional[StrictStr]=None
    sse_customer_key: Optional[StrictStr]=None
    part_number: Optional[StrictInt]=None
    expected_bucket_owner: Optional[StrictStr]=None
    checksum_mode: Optional[
        Literal['ENABLED', 'DISABLED']
    ]=None

    def to_options(self):

        options = self._filtered_options_to_dict()

        sse_options = {
            convert_key_to_boto3_arg_upper_matching(
                key,
                ['sse']
            ): value for key, value in options.items() if key.startswith('sse')
        }

        parsed_options = {
            convert_key_to_boto3_arg(
                key
            ): value for key, value in options.items()
        }

        parsed_options.update(sse_options)

        return parsed_options





import asyncio
import boto3
import functools
from botocore.config import Config
from concurrent.futures import ThreadPoolExecutor
from werkflow_aws.exceptions import UnsetAWSConnectionException
from werkflow_aws.models import (
    AWSCredentialsSet,
    AWSRegionMap,
    RegionName,
)
from werkflow_aws.models.elasticache import (
    ElasticacheDescribeCacheClustersRequest,
    ElasticacheDescribeCacheClustersResponse,
    ElasticacheDescribeReplicationGroupsRequest,
    ElasticacheDescribeReplicationGroupsResponse,
)
from werkflow_aws.helpers import collect_paginator

from werkflow_aws.types import ElastiCacheClient
from werkflow_system import System


class AWSElastiCache:
    
    def __init__(self) -> None:
        self._system = System()

        self._loop: asyncio.AbstractEventLoop | None = None
        self._executor = ThreadPoolExecutor(
            max_workers=self._system.configuration.cores.physical
        )

        self._client = None

        self.service_name = 'ElasticBlockStorage'
        self._regions = AWSRegionMap()
    
        
    async def sso(
        self,
        profile_name: str,
    ):

        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        await self._loop.run_in_executor(
            None,
            functools.partial(
                boto3.setup_default_session,
                profile_name=profile_name
            )
        )

        self._client: ElastiCacheClient = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                'elasticache',
            )
        )

    async def connect(
        self,
        credentials: AWSCredentialsSet,
        region: RegionName,
    ):
        
        aws_region = self._regions.get(region)

        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        self._client: ElastiCacheClient = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                'elasticache',
                aws_access_key_id=credentials.aws_access_key_id,
                aws_secret_access_key=credentials.aws_secret_access_key,
                aws_session_token=credentials.aws_session_token,
                config=Config(
                    region_name=aws_region.value
                )
            )
        )

    async def describe_cache_clusters(
        self,
        request: ElasticacheDescribeCacheClustersRequest
    ) -> ElasticacheDescribeCacheClustersResponse:
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name,
            )
        
        dumped = request.model_dump(exclude_none=True)
        
        result = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.describe_cache_clusters,
                **dumped
            )
        )

        return ElasticacheDescribeCacheClustersResponse(**result)
    

    async def describe_cache_clusters_paginated(
        self,
        request: ElasticacheDescribeCacheClustersRequest
    ) -> list[ElasticacheDescribeCacheClustersResponse]:
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name,
            )
        
        dumped = request.model_dump(exclude_none=True)

        paginator = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.get_paginator,
                'describe_cache_clusters',
            )
        )

        pagination_result = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                paginator.paginate,
                **dumped
            )
        )

        results = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                collect_paginator,
                pagination_result,
            )
        )

        return [
            ElasticacheDescribeCacheClustersResponse(**result) for result in results
        ]
    
    async def describe_replication_groups(
        self,
        request: ElasticacheDescribeReplicationGroupsRequest
    ) -> ElasticacheDescribeReplicationGroupsResponse:
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name,
            )
        
        dumped = request.model_dump(exclude_none=True)
        
        result = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.describe_cache_clusters,
                **dumped
            )
        )

        return ElasticacheDescribeReplicationGroupsResponse(**result)
    
    async def describe_replication_groups_paginated(
        self,
        request: ElasticacheDescribeCacheClustersRequest
    ) -> list[ElasticacheDescribeCacheClustersResponse]:
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name,
            )
        
        dumped = request.model_dump(exclude_none=True)

        paginator = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.get_paginator,
                'describe_replication_groups',
            )
        )

        pagination_result = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                paginator.paginate,
                **dumped
            )
        )

        results = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                collect_paginator,
                pagination_result,
            )
        )

        return [
            ElasticacheDescribeReplicationGroupsResponse(**result) for result in results
        ]
    
    async def close(self):

        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client:
            await self._loop.run_in_executor(
                None,
                self._client.close
            )

        await self._system.close()
        self._executor.shutdown()

    def abort(self):
        self._client.close()
        self._system.abort()
        self._executor.shutdown()


from pydantic import BaseModel, StrictStr, StrictInt
from .policy_arn import PolicyArn
from .provided_context import ProvidedContext
from .tag import Tag


class AssumeRoleRequest(BaseModel):
    RoleArn: StrictStr
    RoleSessionName: StrictStr
    PolicyArns: list[PolicyArn] | None = None
    Policy: StrictStr | None = None
    DurationSeconds: StrictInt | None = None
    Tags: list[Tag] | None = None
    TransitiveTagKeys: list[StrictStr] | None = None
    ExternalId: StrictStr | None = None
    SerialNumber: StrictStr | None = None
    TokenCode: StrictStr | None = None
    SourceIdentity: StrictStr | None = None
    ProvidedContexts: list[ProvidedContext] | None = None


from .organizations_account import OrganizationsAccount as OrganizationsAccount
from .organizations_account_status import AccountStatusName as AccountStatusName
from .organizations_account_status import AWSAccountStatus as AWSAccountStatus
from .organizations_account_status import OrganizationAccountStatusMap as OrganizationAccountStatusMap
from .organizations_account_status import OrganizationsAccountStatus as OrganizationsAccountStatus
from .organizations_joined_method import AWSJoinedMethod as AWSJoinedMethod
from .organizations_joined_method import JoinedMethodName as JoinedMethodName
from .organizations_joined_method import OrganizationsJoinedMethod as OrganizationsJoinedMethod
from .organizations_joined_method import OrganizationsJoinedMethodMap as OrganizationsJoinedMethodMap
from .organizations_list_accounts_request import OrganizationsListAccountsRequest as OrganizationsListAccountsRequest
from .organizations_list_accounts_response import OrganizationsListAccountsResponse as OrganizationsListAccountsResponse


from typing import Any


def collect_paginator(results: list[Any]) -> list[dict[str, Any]]:
    paginated_results: list[dict[str, Any]] = []

    for page in results:
        paginated_results.append(page)

    return paginated_results


import datetime
from pydantic import BaseModel, StrictFloat, StrictStr
from .cloudwatch_statistics import CloudWatchStatistics


class CloudWatchDatapoint(BaseModel):
    Timestamp: datetime.datetime
    SampleCount: StrictFloat | None = None
    Average: StrictFloat | None = None
    Sum: StrictFloat | None = None
    Minimum: StrictFloat | None = None
    Maximum: StrictFloat | None = None
    Unit: CloudWatchStatistics | None = None
    ExtendedStatistics: dict[StrictStr, StrictFloat] | None = None



from pydantic import (
    StrictStr,
    StrictBool
)
from werkflow_aws.models.base import AWSBoto3Base
from typing import Optional, Literal


class AWSs3DeleteObjectsOptions(AWSBoto3Base):
    mfa: Optional[StrictStr]=None
    request_payer: Optional[
        Literal['requester']
    ]=None
    bypass_governance_retention: Optional[StrictBool]=None
    expected_bucket_owner: Optional[StrictStr]=None
    checksum_algorithm: Optional[
        Literal[
            'CRC32',
            'CRC32C',
            'SHA1',
            'SHA256'
        ]
    ]=None


from pydantic import BaseModel, StrictStr
from .cloudwatch_metric_data_result import CloudWatchMetricDataResult
from .cloudwatch_message import CloudWatchMessage

class CloudWatchGetMetricDataResponse(BaseModel):
    MetricDataResults: list[CloudWatchMetricDataResult]
    NextToken: StrictStr | None = None
    Messages: list[CloudWatchMessage] | None = None


from enum import Enum
from typing import Literal


AWSJoinedMethod = Literal[
    "INVITED",
    "CREATED",
]


JoinedMethodName = Literal[
    "invited",
    "created",
]


class OrganizationsJoinedMethod(Enum):
    INVITED='INVITED'
    CREATED='CREATED'


class OrganizationsJoinedMethodMap:

    def __init__(self):
        self._joined_method_map: dict[
            JoinedMethodName,
            OrganizationsJoinedMethod
        ] = {
            'created': OrganizationsJoinedMethod.CREATED,
            'invited': OrganizationsJoinedMethod.INVITED,
        }

        self._joined_method_reverse_map: dict[
            AWSJoinedMethod,
            JoinedMethodName,
        ] = {
            'CREATED': 'created',
            'INVITED': 'invited',
        }

    def get(self, joined_method: JoinedMethodName):
        return self._joined_method_map[joined_method]
    
    def get_joined_method_name(self, joined_method_name: AWSJoinedMethod):
        return self._joined_method_reverse_map[joined_method_name]


import datetime
from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictBool
)
from typing import Literal, Optional, Dict, List
from .aws_s3_part_response import AWSs3PartResponse


class AWSs3ListPartsResponse(BaseModel):
    AbortDate: Optional[datetime.datetime]=None
    AbortRuleId: Optional[StrictStr]=None
    Bucket: StrictStr
    Key: StrictStr
    UploadId: StrictStr
    PartNumberMarker: StrictInt
    NextPartNumberMarker: StrictInt
    MaxParts: StrictInt
    IsTruncated: StrictBool
    Parts: List[AWSs3PartResponse]=[]
    Initiator: Dict[
        Literal['ID', 'DisplayName'],
        StrictStr
    ]
    Owner: Dict[
        Literal['DisplayName', 'ID'],
        StrictStr
    ]
    StorageClass: Literal[
        'STANDARD',
        'REDUCED_REDUNDANCY',
        'STANDARD_IA',
        'ONEZONE_IA',
        'INTELLIGENT_TIERING',
        'GLACIER',
        'DEEP_ARCHIVE',
        'OUTPOSTS',
        'GLACIER_IR',
        'SNOW',
        'EXPRESS_ONEZONE'
    ]
    RequestCharged: Optional[
        Literal['requester']
    ]=None
    ChecksumAlgorithm: Literal[
        'CRC32',
        'CRC32C',
        'SHA1',
        'SHA256'
    ]


import datetime
from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
)

from .msk_client_authentication import MSKClientAuthentication
from .msk_current_broker_software_info import MSKCurrentBrokerSoftwareInfo
from .msk_encryption_info import MSKEncryptionInfo
from .msk_enhanced_monitoring import MSKEnhancedMonitoring
from .msk_logging_info import MSKLoggingInfo
from .msk_open_monitoring import MSKOpenMonitoring
from .msk_node_group_state import MSKNodeGroupState
from .msk_state_info import MSKStateInfo
from .msk_storage_mode import MSKStorageMode
from .msk_customer_action_status import MSKCustomerActionStatus


class MSKCluster(BaseModel):
    ActiveOperationArn: StrictStr
    ClientAuthentication: MSKClientAuthentication
    ClusterArn: StrictStr
    ClusterName: StrictStr
    CreationTime: datetime.datetime
    CurrentBrokerSoftwareInfo: MSKCurrentBrokerSoftwareInfo
    CurrentVersion: StrictStr
    EncryptionInfo: MSKEncryptionInfo
    EnhancedMonitoring: MSKEnhancedMonitoring
    OpenMonitoring: MSKOpenMonitoring
    LoggingInfo: MSKLoggingInfo
    NumberOfBrokerNodes: StrictInt
    State: MSKNodeGroupState
    StateInfo: MSKStateInfo
    Tags: dict[StrictStr, StrictStr] | None = None
    ZookeeperConnectString: StrictStr
    ZookeeperConnectStringTls: StrictStr
    StorageMode: MSKStorageMode
    CustomerActionStatus: MSKCustomerActionStatus



from pydantic import BaseModel
from typing import List
from .aws_s3_bucket_response import AWSs3BucketResponse
from .aws_s3_owner_response import AWSs3OwnerResponse


class AWSs3ListBucketsResponse(BaseModel):
    Buckets: List[AWSs3BucketResponse]=[]
    Owner: List[AWSs3BucketResponse]=[]


from enum import Enum
from typing import Literal


ServiceName = Literal[
    "amplify",
    "api-gateway",
    "app-stream",
    "athena",
    "app-sync",
    "app-flow",
    "app-studio",
    "app-runner",
    "audit-manager",
    "backup",
    "bedrock",
    "cert-manager",
    "codeartifact",
    "cloud-trail",
    "cloud-wan",
    "cloud-formation",
    "cloudfront",
    "cloudhsm",
    "cloudsearch",
    "cloudtrail",
    "cloudwatch",
    "cloudwatch-events",
    "codebuild",
    "codecatalyst",
    "codecommit",
    "codeguru",
    "codepipeline",
    "codewhisperer",
    "cognito",
    "cognito-sync",
    "comprehend",
    "compute-optimizer",
    "config",
    "connect",
    "cost-explorer",
    "data-pipeline",
    "data-transfer",
    "datasync",
    "datazone",
    "deepracer",
    "detective",
    "device-farm",
    "devops-guru",
    "direct-connect",
    "directory-service",
    "dms",
    "documentdb",
    "dynamodb",
    "ec2",
    "ec2-other",
    "ecr",
    "ecr-public",
    "ecs",
    "ecs-kube",
    "efs",
    "elb",
    "emr",
    "elasticache",
    "elemential-mediastore",
    "eum",
    "enterprise-support",
    "fis",
    "fpm",
    "firewall",
    "fsx",
    "glacier",
    "glue",
    "guardduty",
    "healthimaging",
    "iam-analyzer",
    "iot",
    "iot-dm",
    "kendra",
    "kinesis",
    "kinesis-analytics",
    "kinesis-firehose",
    "kinesis-vs",
    "lake-formation",
    "lambda",
    "lex",
    "lightsail",
    "location-service",
    "lookout-metrics",
    "macie",
    "managed-blockchain",
    "managed-grafana",
    "memorydb",
    "migration-hub",
    "msk",
    "mq",
    "mwa",
    "neptune",
    "opensearch-severless",
    "opensearch",
    "personalize",
    "pinpoint",
    "polly",
    "q",
    "quantum-db",
    "quicksight",
    "rds",
    "redshift",
    "refund",
    "registrar",
    "rekognition",
    "resilience-hub",
    "rosa",
    "route53",
    "route53-arc",
    "s3",
    "sagemaker",
    "secrets-manager",
    "security-hub",
    "security-lake",
    "service-catalogue",
    "ses",
    "simpledb",
    "sns",
    "step-functions",
    "storage-gateway",
    "sqs",
    "sumerian",
    "sws",
    "systems-manager",
    "tax",
    "textract",
    "timestream",
    "transcribe",
    "transfer-family",
    "translate",
    "verified-permissions",
    "vpc",
    "waf",
    "workforce-id",
    "workspaces",
    "workspaces-tc",
    "xray",
]

MappedServiceNames = Literal[
    'AWS Amplify',
    'Amazon API Gateway',
    'Amazon AppStream',
    'Amazon Athena',
    'AWS AppSync',
    'Amazon AppFlow',
    'AWS App Studio',
    'AWS App Runner',
    'AWS Audit Manager',
    'AWS Backup',
    'Amazon Bedrock',
    'AWS Certificate Manager',
    'AWS CodeArtifact',
    'AWS CloudTrail',
    'AWS Cloud WAN',
    'AWS CloudFormation',
    'Amazon CloudFront',
    'AWS CloudHSM',
    'Amazon CloudSearch',
    'AWS CloudTrail',
    'AmazonCloudWatch',
    'CloudWatch Events',
    'CodeBuild',
    'CodeCatalyst',
    'AWS CodeCommit',
    'CodeGuru',
    'AWS CodePipeline',
    'Amazon CodeWhisperer',
    'Amazon Cognito',
    'Amazon Cognito Sync',
    'Amazon Comprehend',
    'AWS Compute Optimizer',
    'AWS Config',
    'Amazon Connect',
    'AWS Cost Explorer',
    'AWS Data Pipeline',
    'AWS Data Transfer',
    'AWS DataSync',
    'Amazon DataZone',
    'AWS DeepRacer',
    'Amazon Detective',
    'AWS Device Farm',
    'Amazon DevOps Guru',
    'AWS Direct Connect',
    'AWS Directory Service',
    'AWS Database Migration Service',
    'Amazon DocumentDB (with MongoDB compatibility)',
    'Amazon DynamoDB',
    'Amazon Elastic Compute Cloud - Compute',
    'EC2 - Other',
    'Amazon EC2 Container Registry (ECR)',
    'Amazon Elastic Container Registry Public',
    'Amazon Elastic Container Service',
    'Amazon Elastic Container Service for Kubernetes',
    'Amazon Elastic File System',
    'Amazon Elastic Load Balancing',
    'Amazon Elastic MapReduce',
    'Amazon ElastiCache',
    'AWS Elemental MediaStore',
    'AWS End User Messaging',
    'AWS Support (Enterprise)',
    'AWS Fault Injection Simulator',
    'FinOps Practice Management',
    'AWS Firewall Manager',
    'Amazon FSx',
    'Amazon Glacier',
    'AWS Glue',
    'Amazon GuardDuty',
    'AWS HealthImaging',
    'AWS Identity and Access Management Access Analyzer',
    'AWS IoT',
    'AWS IoT Device Management',
    'Amazon Kendra',
    'Amazon Kinesis',
    'Amazon Kinesis Analytics',
    'Amazon Kinesis Firehose',
    'Amazon Kinesis Video Streams',
    'AWS Lake Formation',
    'AWS Lambda',
    'Amazon Lex',
    'Amazon Lightsail',
    'Amazon Location Service',
    'Amazon Lookout for Metrics',
    'Amazon Macie',
    'Amazon Managed Blockchain',
    'Amazon Managed Grafana',
    'Amazon MemoryDB',
    'AWS Migration Hub Refactor Spaces',
    'Amazon Managed Streaming for Apache Kafka',
    'Amazon MQ',
    'Amazon Managed Workflows for Apache Airflow',
    'Amazon Neptune',
    'Amazon OpenSearch Serverless',
    'Amazon OpenSearch Service',
    'Amazon Personalize',
    'Amazon Pinpoint',
    'Amazon Polly',
    'Amazon Q',
    'Amazon Quantum Ledger Database',
    'Amazon QuickSight',
    'Amazon Relational Database Service',
    'Amazon Redshift',
    'Refund',
    'Amazon Registrar',
    'Amazon Rekognition',
    'AWS Resilience Hub',
    'Red Hat OpenShift Service on AWS',
    'Amazon Route 53',
    'AWS Route 53 Application Recovery Controller',
    'Amazon Simple Storage Service',
    'Amazon SageMaker',
    'AWS Secrets Manager',
    'AWS Security Hub',
    'Amazon Security Lake',
    'AWS Service Catalog',
    'Amazon Simple Email Service',
    'Amazon SimpleDB',
    'Amazon Simple Notification Service',
    'AWS Step Functions',
    'AWS Storage Gateway',
    'Amazon Simple Queue Service',
    'Amazon Sumerian',
    'Amazon Simple Workflow Service',
    'AWS Systems Manager',
    'Tax',
    'Amazon Textract',
    'Amazon Timestream',
    'Amazon Transcribe',
    'AWS Transfer Family',
    'Amazon Translate',
    'Amazon Verified Permissions',
    'Amazon Virtual Private Cloud',
    'AWS WAF',
    'Workforce Identity Solutions',
    'Amazon WorkSpaces',
    'Amazon WorkSpaces Thin Client',
    'AWS X-Ray'
]


class AWSServices(Enum):
    AMPLIFY='AWS Amplify'
    API_GATEWAY="Amazon API Gateway"
    APP_STREAM="Amazon AppStream"
    ATHENA="Amazon Athena"
    APP_SYNC="AWS AppSync"
    APP_FLOW="Amazon AppFlow"
    APP_STUDIO="AWS App Studio"
    APP_RUNNER="AWS App Runner"
    AUDIT_MANAGER="AWS Audit Manager"
    BACKUP="AWS Backup"
    BEDROCK="Amazon Bedrock"
    CERT_MANAGER="AWS Certificate Manager"
    CODEARTIFACT="AWS CodeArtifact"
    CLOUD_TRAIL="AWS CloudTrail"
    CLOUD_WAN="AWS Cloud WAN"
    CLOUD_FORMATION="AWS CloudFormation"
    CLOUDFRONT="Amazon CloudFront"
    CLOUDHSM="AWS CloudHSM"
    CLOUDSEARCH="Amazon CloudSearch"
    CLOUDTRAIL="AWS CloudTrail"
    CLOUDWATCH="AmazonCloudWatch"
    CLOUDWATCH_EVENTS="CloudWatch Events"
    CODEBUILD="CodeBuild"
    CODECATALYST="CodeCatalyst"
    CODECOMMIT="AWS CodeCommit"
    CODEGURU="CodeGuru"
    CODEPIPELINE="AWS CodePipeline"
    CODEWHISPERER="Amazon CodeWhisperer"
    COGNITO="Amazon Cognito"
    COGNITO_SYNC="Amazon Cognito Sync"
    COMPREHEND="Amazon Comprehend"
    COMPUTE_OPTIMIZER="AWS Compute Optimizer"
    CONFIG="AWS Config"
    CONNECT="Amazon Connect"
    COST_EXPLORER="AWS Cost Explorer"
    DATA_PIPELINE="AWS Data Pipeline"
    DATA_TRANSFER="AWS Data Transfer"
    DATASYNC="AWS DataSync"
    DATAZONE="Amazon DataZone"
    DEEPRACER="AWS DeepRacer"
    DETECTIVE="Amazon Detective"
    DEVICE_FARM="AWS Device Farm"
    DEVOPS_GURU="Amazon DevOps Guru"
    DIRECT_CONNECT="AWS Direct Connect"
    DIRECTORY_SERVICE="AWS Directory Service"
    DMS="AWS Database Migration Service"
    DOCUMENTDB="Amazon DocumentDB (with MongoDB compatibility)"
    DYNAMODB="Amazon DynamoDB"
    EC2='Amazon Elastic Compute Cloud - Compute'
    EC2_OTHER='EC2 - Other'
    ECR='Amazon EC2 Container Registry (ECR)'
    ECR_PUBLIC='Amazon Elastic Container Registry Public'
    ECS='Amazon Elastic Container Service'
    ECS_KUBE='Amazon Elastic Container Service for Kubernetes'
    EFS='Amazon Elastic File System'
    ELB='Amazon Elastic Load Balancing'
    EMR='Amazon Elastic MapReduce'
    ELASTICACHE='Amazon ElastiCache'
    ELEMENTAL_MEDIASTORE='AWS Elemental MediaStore'
    EUM='AWS End User Messaging'
    ENTERPRISE_SUPPORT='AWS Support (Enterprise)'
    FIS='AWS Fault Injection Simulator'
    FPM='FinOps Practice Management'
    FIREWALL='AWS Firewall Manager'
    FSX='Amazon FSx'
    GLACIER='Amazon Glacier'
    GLUE='AWS Glue'
    GUARDDUTY='Amazon GuardDuty'
    HEALTHIMAGING='AWS HealthImaging'
    IAM_ANALYZER='AWS Identity and Access Management Access Analyzer'
    IOT='AWS IoT'
    IOT_DM='AWS IoT Device Management'
    KENDRA='Amazon Kendra'
    KINESIS='Amazon Kinesis'
    KINESIS_ANALYTICS='Amazon Kinesis Analytics'
    KINESIS_FIREHOSE='Amazon Kinesis Firehose'
    KINESIS_VS='Amazon Kinesis Video Streams'
    LAKE_FORMATION='AWS Lake Formation'
    LAMBDA='AWS Lambda'
    LEX='Amazon Lex'
    LIGHTSAIL='Amazon Lightsail'
    LOCATION_SERVICE='Amazon Location Service'
    LOOKOUT_METRICS='Amazon Lookout for Metrics'
    MACIE='Amazon Macie'
    MANAGED_BLOCKCHAIN='Amazon Managed Blockchain'
    MANAGED_GRAFANA='Amazon Managed Grafana'
    MEMORYDB='Amazon MemoryDB'
    MIGRATION_HUB='AWS Migration Hub Refactor Spaces'
    MSK='Amazon Managed Streaming for Apache Kafka'
    MQ='Amazon MQ'
    MWA='Amazon Managed Workflows for Apache Airflow'
    NEPTUNE='Amazon Neptune'
    OPENSEARCH_SEVERLESS='Amazon OpenSearch Serverless'
    OPENSEARCH='Amazon OpenSearch Service'
    PERSONALIZE='Amazon Personalize'
    PINPOINT='Amazon Pinpoint'
    POLLY='Amazon Polly'
    Q='Amazon Q'
    QUANTUM_DB='Amazon Quantum Ledger Database'
    QUICKSIGHT='Amazon QuickSight'
    RDS='Amazon Relational Database Service'
    REDSHIFT='Amazon Redshift'
    REFUND='Refund'
    REGISTRAR='Amazon Registrar'
    REKOGNITION='Amazon Rekognition'
    RESILIENCE_HUB='AWS Resilience Hub'
    ROSA='Red Hat OpenShift Service on AWS'
    ROUTE53='Amazon Route 53'
    ROUTE53_ARC='AWS Route 53 Application Recovery Controller'
    S3='Amazon Simple Storage Service'
    SAGEMAKER='Amazon SageMaker'
    SECRETS_MANAGER='AWS Secrets Manager'
    SECURITY_HUB='AWS Security Hub'
    SECURITY_LAKE='Amazon Security Lake'
    SERVICE_CATALOGUE='AWS Service Catalog'
    SES='Amazon Simple Email Service'
    SIMPLEDB='Amazon SimpleDB'
    SNS='Amazon Simple Notification Service'
    STEP_FUNCTIONS='AWS Step Functions'
    STORAGE_GATEWAY='AWS Storage Gateway'
    SQS='Amazon Simple Queue Service'
    SUMERIAN='Amazon Sumerian'
    SWS='Amazon Simple Workflow Service'
    SYSTEMS_MANAGER='AWS Systems Manager'
    TAX='Tax'
    TEXTRACT='Amazon Textract'
    TIMESTREAM='Amazon Timestream'
    TRANSCRIBE='Amazon Transcribe'
    TRANSFER_FAMILY='AWS Transfer Family'
    TRANSLATE='Amazon Translate'
    VERIFIED_PERMISSIONS='Amazon Verified Permissions'
    VPC='Amazon Virtual Private Cloud'
    WAF='AWS WAF'
    WORKFORCE_ID='Workforce Identity Solutions'
    WORKSPACES='Amazon WorkSpaces'
    WORKSPACES_TC='Amazon WorkSpaces Thin Client'
    XRAY='AWS X-Ray'


class AWSServicesMap:
    
    def __init__(self):
        self._services_map: dict[
            ServiceName,
            MappedServiceNames,
        ] = {
            service.name.lower().replace(
                '_',
                '-'
            ): service.value for service in AWSServices
        }

        self._service_to_name_map: dict[
            AWSServices,
            ServiceName
        ] = {
            value: key for key, value in self._services_map.items()
        }

    def get(self, service_name: ServiceName):
        return self._services_map[service_name]
    
    def get_service_name(self, service: MappedServiceNames):
        return self._service_to_name_map[service]


from pydantic import BaseModel, StrictStr, StrictBool, StrictInt
from .cloudwatch_metric_stat import CloudWatchMetricStat

class CloudWatchMetricDataQuery(BaseModel):
    Id: StrictStr
    MetricStat: CloudWatchMetricStat
    Expression: StrictStr | None = None
    Label: StrictStr | None = None
    ReturnData: StrictBool = False
    Period: StrictInt | None = None
    AccountId: StrictStr | None = None


from enum import Enum
from typing import Literal

AWSAccountStatus = Literal[
    "ACTIVE",
    "SUSPENDED",
    "PENDING_CLOSURE",
]

AccountStatusName = Literal[
    "active",
    "suspended",
    "pending-closure"
]


class OrganizationsAccountStatus(Enum):
    ACTIVE="ACTIVE"
    SUSPENDED="SUSPENDED"
    PENDING_CLOSURE="PENDING_CLOSURE"


class OrganizationAccountStatusMap:

    def __init__(self):
        self._status_map: dict[
            AccountStatusName,
            OrganizationsAccountStatus,
        ] = {
            'active': OrganizationsAccountStatus.ACTIVE,
            'pending-closure': OrganizationsAccountStatus.PENDING_CLOSURE,
            'suspended': OrganizationsAccountStatus.SUSPENDED,
        }

        self._reverse_status_map: dict[
            AWSAccountStatus,
            AccountStatusName
        ] = {
            'ACTIVE': 'active',
            'PENDING_CLOSURE': 'pending-closure',
            'SUSPENDED': 'suspended'
        }

    def get(self, account_status: AccountStatusName):
        return self._status_map[account_status]
    
    def get_status_name(self, aws_account_status: AWSAccountStatus):
        return self._reverse_status_map[aws_account_status]


from .aws_credentials_set import AWSCredentialsSet as AWSCredentialsSet
from .aws_region import AWSRegion as AWSRegion
from .aws_region import AWSRegionMap as AWSRegionMap
from .aws_region import RegionName as RegionName
from .aws_services import AWSServices as AWSServices
from .aws_services import AWSServicesMap as AWSServicesMap
from .aws_services import MappedServiceNames as MappedServiceNames
from .aws_services import ServiceName as ServiceName


from __future__ import annotations

from botocore.paginate import Paginator
from abc import ABC, abstractmethod
from typing import Any


class MSKClient(ABC):

    @abstractmethod
    def list_clusters(
        self,     
        ClusterNameFilter: str | None = None,
        MaxResults: int | None = None,
        NextToken: str | None = None,
    ) -> dict[str, Any]:
        pass

    @abstractmethod
    def get_paginator(
        self,
        Operation: str,
    ) -> Paginator:
        pass

    @abstractmethod
    def close(self):
        pass


# response = client.list_clusters(
#     ClusterNameFilter='string',
#     MaxResults=123,
#     NextToken='string'
# )

from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
)


class MSKListClustersRequest(BaseModel):
    ClusterNameFilter: StrictStr | None = None
    MaxResults: StrictInt | None = None
    NextToken: StrictStr | None = None


from pydantic import (
    StrictBool,
    StrictStr,
    StrictInt
)

from typing import Optional, List
from werkflow_aws.models.base import AWSBoto3Base


class AWSs3ListObjectsOptions(AWSBoto3Base):
    delimiter: Optional[StrictStr]=None
    encoding_type: Optional[StrictStr]=None
    max_keys: Optional[StrictInt]=None
    prefix: Optional[StrictStr]=None
    continuation_token: Optional[StrictStr]=None
    fetch_owner: Optional[StrictBool]=None
    start_after: Optional[StrictStr]=None
    expected_bucket_owner: Optional[StrictStr]=None
    optional_object_attributes: Optional[List[StrictStr]]=None




import asyncio
import functools
from concurrent.futures import ThreadPoolExecutor

import boto3
from botocore.config import Config
from werkflow_aws.exceptions import UnsetAWSConnectionException
from werkflow_aws.models import (
    AWSCredentialsSet,
    AWSRegionMap,
    RegionName,
)
from werkflow_aws.models.sts import (
    AssumeRoleRequest,
    AssumedRoleResponse,
)
from werkflow_system import System
from werkflow_aws.types import STSClient


class AWSSTS:

    def __init__(self):
        self._system = System()

        self._loop: asyncio.AbstractEventLoop | None = None
        self._executor = ThreadPoolExecutor(
            max_workers=self._system.configuration.cores.physical
        )

        self._client = None

        self.service_name = 'STS'

        self._regions = AWSRegionMap()

    async def connect(
        self,
        credentials: AWSCredentialsSet,
        region: RegionName,
    ):
        
        aws_region = self._regions.get(region)
        
        if self._loop is None:
            self._loop = asyncio.get_event_loop()       

        self._client: STSClient = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                'sts',
                aws_access_key_id=credentials.aws_access_key_id,
                aws_secret_access_key=credentials.aws_secret_access_key,
                aws_session_token=credentials.aws_session_token,
                config=Config(
                    region_name=aws_region.value
                )
            )
        )

    async def connect_unauthorized(
        self,
        region: RegionName,
    ):
        
        aws_region = self._regions.get(region)

        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        self._client: STSClient = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                'sts',
                config=Config(
                    region_name=aws_region.value
                )
            )
        )

    async def assume_role(
        self,
        assume_role_request: AssumeRoleRequest
    ):
        
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            )

        response = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.assume_role,
                **assume_role_request.model_dump(exclude_none=True)
            )
        )

        return AssumedRoleResponse(**response)
    
    async def close(self):
        
        if self._loop is None:
            self._loop = asyncio.get_event_loop()     

        if self._client:
            await self._loop.run_in_executor(
                None,
                self._client.close
            )

        await self._system.close()
        self._executor.shutdown()

    def abort(self):
        self._client.close()
        self._system.abort()
        self._executor.shutdown()


from enum import Enum
from typing import Literal


GranularityLevel = Literal['hourly', 'daily', 'monthly']
AWSGranularityLevel = Literal['HOURLY', 'DAILY', 'MONTHLY']


class Granularity(Enum):
    HOURLY='HOURLY'
    DAILY='DAILY'
    MONTHLY='MONTHLY'

    @classmethod
    def to_granularity(cls, granularity: GranularityLevel):
        levels = {
            'hourly': Granularity.HOURLY,
            'daily': Granularity.DAILY,
            'monthly': Granularity.MONTHLY
        }

        return levels.get(granularity, Granularity.DAILY)

    @classmethod
    def to_aws_granularity_slug(cls, granularity: GranularityLevel):
        levels: dict[GranularityLevel, AWSGranularityLevel] = {
            level.name.lower(): level.value for level in Granularity
        }

        return levels.get(granularity, Granularity.DAILY.value)


from pydantic import BaseModel, StrictStr
from .elasticache_cache_cluster import ElasticacheCacheCluster


class ElasticacheDescribeCacheClustersResponse(BaseModel):
    Marker: StrictStr
    CacheClusters: list[ElasticacheCacheCluster]




from __future__ import annotations
from pydantic import BaseModel, StrictStr
from typing import Literal
from .dump_options import DumpOptions
from .expression import Expression
from .group import Group
from .granularity import AWSGranularityLevel
from .time_period import TimePeriod


ValidMetrics = Literal[
    "AmortizedCost",
    "BlendedCost",
    "NetAmortizedCost",
    "NetUnblendedCost",
    "NormalizedUsageAmount",
    "UnblendedCost",
    "UsageQuantity",
]



class CostExplorerQuery(BaseModel):
    TimePeriod: TimePeriod
    Granularity: AWSGranularityLevel = 'MONTHLY'
    Filter: Expression | None = None
    Metrics: list[ValidMetrics]
    GroupBy: list[Group] | None = None
    BillingViewArn: StrictStr | None = None
    NextPageToken: StrictStr | None = None

    def dump(self, options: DumpOptions | None = None):
        if options is None:
            options = DumpOptions(
                time_format='%Y-%m-%d',
            )

        dumped = {
            'TimePeriod': self.TimePeriod.dump(options.time_format),
            'Granularity': self.Granularity,
            'Metrics': self.Metrics,
        }

        if filters := self.Filter:
            dumped['Filter'] = filters.dump()

        if groups := self.GroupBy:
            dumped['GroupBy'] = [
                group.dump() for group in groups
            ]

        if self.BillingViewArn:
            dumped['BillingViewArn'] = self.BillingViewArn

        if self.NextPageToken:
            dumped['NextPageToken'] = self.NextPageToken

        return dumped


import asyncio
import boto3
import functools
from botocore.config import Config
from concurrent.futures import ThreadPoolExecutor
from io import FileIO
from werkflow_aws.exceptions import UnsetAWSConnectionException
from werkflow_aws.models import (
    AWSCredentialsSet,
    AWSRegionMap,
    RegionName,
)
from werkflow_aws.models.s3 import (
    AWSs3AbortMultipartUploadOptions,
    AWSs3AbortMultipartUploadResponse,
    AWSs3CreateMultipartUploadOptions,
    AWSs3CreateMultipartUploadResponse,
    AWSs3CompleteMultipartUploadOptions,
    AWSs3CompleteMultipartUploadResponse,
    AWSs3CopySource,
    AWSs3CopyObjectOptions,
    AWSs3CopyObjectResponse,
    AWSs3DeleteBucketOptions,
    AWSs3DeleteBucketReponse,
    AWSs3DeleteBucketTaggingOptions,
    AWSs3DeleteObjectTaggingResponse,
    AWSs3DeleteObjectOptions,
    AWSs3DeleteObjectResponse,
    AWSs3DeleteObjectTaggingOptions,
    AWSs3DeleteBucketTaggingResponse,
    AWSs3DeleteObjectsOptions,
    AWSs3DeleteObjectsResponse,
    AWSs3GeneratePresignedURLOptions,
    AWSs3GeneratePresignedURLResponse,
    AWSs3GetBucketTaggingOptions,
    AWSs3GetBucketTaggingResponse,
    AWSs3GetObjectOptions,
    AWSs3GetObjectResponse,
    AWSs3GetObjectTaggingOptions,
    AWSs3GetObjectTaggingResponse,
    AWSs3ListBucketsResponse,
    AWSs3GeneratePresignedPostOptions,
    AWSs3GeneratePresignedPostResponse,
    AWSs3ListDirectoryBucketsOptions,
    AWSs3ListDirectoryBucketsResponse,
    AWSs3ListObjectsOptions,
    AWSs3ListObjectsResponse,
    AWSs3ListObjectVersionsOptions,
    AWSs3ListObjectVersionsResponse,
    AWSs3ListMultipartUploadsOptions,
    AWSs3ListMultipartUploadsResponse,
    AWSs3ListPartsOptions,
    AWSs3ListPartsResponse,
    AWSs3MultipartUpload,
    AWSs3Object,
    AWSs3PutBucketTaggingOptions,
    AWSs3PutBucketTaggingResponse,
    AWSs3PutBucketVersioningOptions,
    AWSs3PutBucketVersioningResponse,
    AWSs3PutObjectOptions,
    AWSs3PutObjectResponse,
    AWSs3PutObjectTaggingOptions,
    AWSs3PutObjectTaggingResponse,
    AWSs3StreamingBody,
    AWSs3Tag,
    AWSs3TransferConfigOptions,
    AWSs3TransferAllowedUploadArgs,
    AWSs3UploadPartCopyOptions,
    AWSs3UploadPartCopyResponse,
    AWSs3UploadPartOptions,
    AWSs3UploadPartResponse
)
from typing import List


from werkflow_aws.types import (
    s3Client
)
from werkflow_system import System
from typing import (
    Union, 
    Optional,
    BinaryIO,
    TextIO
)


class AWSs3:
    
    def __init__(self) -> None:
        
        self._system = System()

        self._loop: Union[asyncio.AbstractEventLoop, None] = None
        self._executor = ThreadPoolExecutor(
            max_workers=self._system.configuration.cores.physical
        )

        self._client = None

        self.service_name = 's3'
        self._regions = AWSRegionMap()
    
    async def sso(
        self,
        profile_name: str,
    ):

        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        await self._loop.run_in_executor(
            None,
            functools.partial(
                boto3.setup_default_session,
                profile_name=profile_name
            )
        )

        self._client: s3Client = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                's3',
            )
        )

    async def connect(
        self,
        credentials: AWSCredentialsSet,
        region: RegionName,
    ):

        aws_region = self._regions.get(region)

        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        self._client: s3Client = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                boto3.client,
                's3',
                aws_access_key_id=credentials.aws_access_key_id,
                aws_secret_access_key=credentials.aws_secret_access_key,
                aws_session_token=credentials.aws_session_token,
                config=Config(
                    region_name=aws_region.value
                )
            )
        )

    async def list_buckets(self):

        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        
        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        listed_buckets = await self._loop.run_in_executor(
            self._executor,
            self._client.list_buckets
        )
        
        return AWSs3ListBucketsResponse(**listed_buckets)

    async def list_objects(
        self,
        bucket: str,
        options: AWSs3ListObjectsOptions | None=None
    ):
        
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 
        
        if options is None:
            options = AWSs3ListObjectsOptions()

        listed_objects = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.list_objects_v2,
                Bucket=bucket,
                **options.to_data()
            )
        )

        return AWSs3ListObjectsResponse(**listed_objects)

    async def get_object(
        self,
        bucket: str,
        key: str,
        options: AWSs3GetObjectOptions | None=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 
        
        if options is None:
            options = AWSs3GetObjectOptions()

        retrieved_object = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.get_object,
                Bucket=bucket,
                Key=key,
                **options.to_data()
            )
        )

        return AWSs3GetObjectResponse(
            **retrieved_object,
            Body=AWSs3StreamingBody(
                retrieved_object['Body']
            )
        )

    async def put_object(
        self,
        bucket: str,
        key: str,
        body: bytes | BinaryIO | TextIO,
        options: AWSs3PutObjectOptions | None=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 
        
        if options is None:
            options = AWSs3PutObjectOptions()

        updated_object = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.put_object,
                Bucket=bucket,
                Key=key,
                Body=body,
                **options.to_data()
            )
        )

        return AWSs3PutObjectResponse(**updated_object)

    async def delete_bucket(
            self,
            bucket: str,
            options: Optional[AWSs3DeleteBucketOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3DeleteBucketOptions()

        deleted_bucket = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.delete_bucket,
                Bucket=bucket,
                **options.to_data()
            )
        )

        return AWSs3DeleteBucketReponse(**deleted_bucket)
    
    async def delete_object(
        self,
        bucket: str,
        key: str,
        options: Optional[AWSs3DeleteObjectOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 
        
        if options is None:
            options = AWSs3DeleteObjectOptions()

        deleted_object = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.delete_bucket,
                Bucket=bucket,
                Key=key,
                **options.to_data()
            )
        )
        
        return AWSs3DeleteObjectResponse(**deleted_object)

    async def list_multipart_uploads(
        self,
        bucket: str,
        options: Optional[AWSs3ListMultipartUploadsOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 
        
        if options is None:
            options = AWSs3ListMultipartUploadsOptions()

        multipart_uploads = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.list_multipart_uploads,
                Bucket=bucket,
                **options.to_data()
            )
        )

        return AWSs3ListMultipartUploadsResponse(**multipart_uploads)

    async def list_multipart_upload_parts(
        self,
        bucket: str,
        key: str,
        options: Optional[AWSs3ListPartsOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3ListPartsOptions()

        parts = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.list_parts,
                Bucket=bucket,
                **options.to_data()
            )
        )

        return AWSs3ListPartsResponse(**parts)

    async def complete_multipart_upload(
        self,
        bucket: str,
        key: str,
        upload_parts: AWSs3MultipartUpload,
        options: Optional[AWSs3CompleteMultipartUploadOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3CompleteMultipartUploadOptions()

        completed_multipart_upload = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.complete_multipart_upload,
                Bucket=bucket,
                Key=key,
                MultipartUpload=upload_parts.to_data(),
                **options.to_data()
            )
        )

        return AWSs3CompleteMultipartUploadResponse(**completed_multipart_upload)

    async def abort_multipart_upload(
        self,
        bucket: str,
        key: str,
        upload_id: str,
        options: Optional[AWSs3AbortMultipartUploadOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3AbortMultipartUploadOptions()
        
        aborted_uploads = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.abort_multipart_upload,
                Bucket=bucket,
                Key=key,
                UploadId=upload_id,
                **options.to_data()
            )
        )

        return AWSs3AbortMultipartUploadResponse(**aborted_uploads)

    async def create_multipart_upload(
        self,
        bucket: str,
        key: str,
        options: Optional[AWSs3CreateMultipartUploadOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3CreateMultipartUploadOptions()

        created_multipart_upload = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.create_multipart_upload,
                Bucket=bucket,
                Key=key,
                **options.to_data()
            )
        )

        return AWSs3CreateMultipartUploadResponse(**created_multipart_upload)

    async def upload_multiupload_part(
        self,
        bucket: str,
        key: str,
        part_number: str,
        upload_id: str,
        options: AWSs3UploadPartOptions | None = None
    ): 
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3UploadPartOptions()

        uploaded_part = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.upload_part,
                Bucket=bucket,
                Key=key,
                PartNumber=part_number,
                UploadId=upload_id,
                **options.to_data()
            )
        )

        return AWSs3UploadPartResponse(**uploaded_part)

    async def upload_multiupload_part_copy(
        self,
        bucket: str,
        key: str,
        part_number: str,
        upload_id: str,
        options: AWSs3UploadPartCopyOptions | None = None,
    ): 
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3UploadPartCopyOptions()

        copied_upload_part = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.upload_part_copy,
                Bucket=bucket,
                Key=key,
                PartNumber=part_number,
                UploadId=upload_id,
                **options.to_data()
            )
        )

        return AWSs3UploadPartCopyResponse(**copied_upload_part)
    
    async def upload_file_object(
        self,
        file_object: FileIO,
        bucket: str,
        key: str,
        options: Optional[AWSs3TransferAllowedUploadArgs]=None,
        config: Optional[AWSs3TransferConfigOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3TransferAllowedUploadArgs()

        if config is None:
            config = AWSs3TransferConfigOptions()

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.upload_fileobj,
                file_object,
                bucket,
                key,
                options,
                None,
                config
            )
        )

        await self._loop.run_in_executor(
            self._executor,
            file_object.close
        )

    async def upload_file(
        self,
        path: str,
        bucket: str,
        key: str,
        options: Optional[AWSs3TransferAllowedUploadArgs]=None,
        config: Optional[AWSs3TransferConfigOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3TransferAllowedUploadArgs()

        if config is None:
            config = AWSs3TransferConfigOptions()

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.upload_file,
                path,
                bucket,
                key,
                options,
                None,
                config
            )
        )

    async def list_directory_buckets(
        self,
        options: Optional[AWSs3ListDirectoryBucketsOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3ListDirectoryBucketsOptions()

        directory_buckets = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.list_directory_buckets,
                **options.to_data()
            )
        )

        return AWSs3ListDirectoryBucketsResponse(**directory_buckets)
    
    async def list_object_versions(
        self,
        bucket: str=None,
        options: Optional[AWSs3ListObjectVersionsOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3ListObjectVersionsOptions()

        directory_buckets = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.list_object_versions,
                Bucket=bucket,
                **options.to_data()
            )
        )

        return AWSs3ListObjectVersionsResponse(**directory_buckets)
    
    async def delete_objects(
        self,
        bucket: str,
        objects: List[AWSs3Object],
        quiet: bool=False,
        options: Optional[AWSs3DeleteObjectsOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3DeleteObjectsOptions()

        delete_request = {
            'Delete': [
                s3_object.model_dump() for s3_object in objects
            ],
            'Quiet': quiet
        }

        deleted_objects = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.list_object_versions,
                Bucket=bucket,
                Delete=delete_request,
                **options.to_data()
            )
        )

        return AWSs3DeleteObjectsResponse(**deleted_objects)
    
    async def download_file_object(
        self,
        file_object: FileIO,
        bucket: str,
        key: str,
        options: Optional[AWSs3TransferAllowedUploadArgs]=None,
        config: Optional[AWSs3TransferConfigOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3TransferAllowedUploadArgs()

        if config is None:
            config = AWSs3TransferConfigOptions()

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.download_fileobj,
                file_object,
                bucket,
                key,
                options,
                None,
                config
            )
        )

        await self._loop.run_in_executor(
            self._executor,
            file_object.close
        )

    async def download_file(
        self,
        path: str,
        bucket: str,
        key: str,
        options: Optional[AWSs3TransferAllowedUploadArgs]=None,
        config: Optional[AWSs3TransferConfigOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3TransferAllowedUploadArgs()

        if config is None:
            config = AWSs3TransferConfigOptions()

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.download_file,
                path,
                bucket,
                key,
                options,
                None,
                config
            )
        )
    
    async def generate_presigned_url(
        self,
        client_method: str,
        options: Optional[AWSs3GeneratePresignedURLOptions]=None
    ) -> str:
        
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3GeneratePresignedURLOptions()

        presigned_url = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.generate_presigned_url,
                client_method,
                **options.to_data()
            )
        )

        return AWSs3GeneratePresignedURLResponse(
            URL=presigned_url
        )
    
    async def geneate_presigned_post(
        self,
        bucket: str,
        key: str,
        options: Optional[AWSs3GeneratePresignedPostOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3GeneratePresignedPostOptions()

        presigned_post = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.generate_presigned_post,
                Bucket=bucket,
                Key=key,
                **options.to_data()
            )
        )

        return AWSs3GeneratePresignedPostResponse(**presigned_post)
    
    async def copy_object(
        self,
        bucket: str,
        copy_source: str | AWSs3CopySource,
        key: str,
        options: Optional[AWSs3CopyObjectOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3CopyObjectOptions()

        if isinstance(copy_source, AWSs3CopySource):
            copy_source = copy_source.to_data()

        copied_object = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.copy_object,
                Bucket=bucket,
                CopySource=copy_source,
                Key=key,
                **options.to_data()
            )
        )

        return AWSs3CopyObjectResponse(**copied_object)
    
    async def copy(
        self,
        copy_source: AWSs3CopySource,
        bucket: str,
        key: str,
        config: Optional[AWSs3TransferConfigOptions]=None,
        options: AWSs3TransferAllowedUploadArgs | None = None,
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3TransferAllowedUploadArgs()

        if config is None:
            config = AWSs3TransferConfigOptions()

        await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.copy,
                copy_source.to_data(),
                bucket,
                key,
                ExtraArgs=options,
                Callback=None,
                SourceClient=None,
                Config=config
            )
        )

    async def put_bucket_versioning(
        self,
        bucket: str,
        options: Optional[AWSs3PutBucketVersioningOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3PutBucketVersioningOptions()

        bucket_versioning = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.put_bucket_versioning,
                Bucket=bucket,
                **options.to_data()
            )
        )

        return AWSs3PutBucketVersioningResponse(**bucket_versioning)
    
    async def get_bucket_tagging(
        self,
        bucket: str,
        options: Optional[AWSs3GetBucketTaggingOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3GetBucketTaggingOptions()

        bucket_tagging = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.put_bucket_tagging,
                Bucket=bucket,
                **options.to_data()
            )
        )

        return AWSs3GetBucketTaggingOptions(**bucket_tagging)
    
    async def get_object_tagging(
        self,
        bucket: str,
        key: str,
        options: Optional[AWSs3GetObjectTaggingOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3GetObjectTaggingOptions()

        bucket_tagging = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.put_bucket_tagging,
                Bucket=bucket,
                Key=key,
                **options.to_data()
            )
        )

        return AWSs3GetObjectTaggingResponse(**bucket_tagging)
    
    async def put_bucket_tagging(
        self,
        bucket: str,
        tags: List[AWSs3Tag],
        options: Optional[AWSs3PutBucketTaggingOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3PutBucketTaggingOptions()

        bucket_tagging = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.put_bucket_tagging,
                Bucket=bucket,
                Tagging={
                    'TagSet': [
                        tag.to_data() for tag in tags
                    ]
                },
                **options.to_data()
            )
        )

        return AWSs3PutBucketTaggingResponse(**bucket_tagging)
    
    async def put_bucket_tagging(
        self,
        bucket: str,
        key: str,
        tags: List[AWSs3Tag],
        options: Optional[AWSs3PutObjectTaggingOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3PutObjectTaggingOptions()

        object_tagging = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.put_object_tagging,
                Bucket=bucket,
                Key=key,
                Tagging={
                    'TagSet': [
                        tag.to_data() for tag in tags
                    ]
                },
                **options.to_data()
            )
        )

        return AWSs3PutObjectTaggingResponse(**object_tagging)
    
    async def delete_bucket_tagging(
        self,
        bucket: str,
        options: Optional[AWSs3DeleteBucketTaggingOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3PutObjectTaggingOptions()

        deleted_bucket_tagging = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.delete_bucket_tagging,
                Bucket=bucket,
                **options.to_data()
            )
        )

        return AWSs3DeleteBucketTaggingResponse(**deleted_bucket_tagging)
    
    async def delete_object_tagging(
        self,
        bucket: str,
        key: str,
        options: Optional[AWSs3DeleteObjectTaggingOptions]=None
    ):
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        if self._client is None:
            raise UnsetAWSConnectionException(
                self.service_name
            ) 

        if options is None:
            options = AWSs3DeleteObjectTaggingOptions()

        deleted_object_tagging = await self._loop.run_in_executor(
            self._executor,
            functools.partial(
                self._client.delete_object_tagging,
                Bucket=bucket,
                Key=key,
                **options.to_data()
            )
        )

        return AWSs3DeleteObjectTaggingResponse(**deleted_object_tagging)

    async def close(self):

        if self._loop is None:
            self._loop = asyncio.get_event_loop()     

        if self._client:
            await self._loop.run_in_executor(
                None,
                self._client.close
            )

        await self._system.close()
        self._executor.shutdown()

    def abort(self):
        self._client.close()
        self._system.abort()
        self._executor.shutdown()


from pydantic import BaseModel, StrictStr
from .cloudwatch_metric import CloudWatchMetric


class CloudWatchListMetricsResponse(BaseModel):
    Metrics: list[CloudWatchMetric]
    NextToken: StrictStr | None = None
    OwningAccounts: list[StrictStr] | None = None


import datetime
from pydantic import BaseModel, StrictStr, StrictInt, StrictBool
from typing import Literal
from .elasticache_endpoint import ElasticacheEndpoint
from .elasticache_log_delivery_configuration import ElasticacheLogDeliveryConfiguration
from .elasticache_pending_modified_values import ElasticachePendingModifiedValues
from .elasticache_notification_configuration import ElasticacheNotificationConfiguration
from .elasticache_cache_secruity_group import ElasticacheCacheSecurityGroup
from .elasticache_cache_parameter_group import ElasticacheCacheParameterGroup
from .elasticache_cache_node import ElasticacheCacheNode
from .elasticache_security_group import ElasticacheSecurityGroup


class ElasticacheCacheCluster(BaseModel):
    CacheClusterId: StrictStr
    ConfigurationEndpoint: ElasticacheEndpoint
    ClientDownloadLandingPage: StrictStr
    Engine: StrictStr
    EngineVersion: StrictStr
    CacheClusterStatus: StrictStr
    NumCacheNodes: StrictInt
    PreferredAvailabilityZone: StrictStr
    PreferredOutpostArn: StrictStr
    CacheClusterCreateTime: datetime.datetime
    PreferredMaintenanceWindow: StrictStr
    PendingModifiedValues: ElasticachePendingModifiedValues
    NotificationConfiguration: ElasticacheNotificationConfiguration
    CacheSecurityGroups: list[ElasticacheCacheSecurityGroup]
    CacheParameterGroup: ElasticacheCacheParameterGroup
    CacheSubnetGroupName: StrictStr
    CacheNodes: list[ElasticacheCacheNode]
    AutoMinorVersionUpgrade: StrictBool
    SecurityGroups: list[ElasticacheSecurityGroup]
    ReplicationGroupId: StrictStr
    SnapshotRetentionLimit: StrictInt
    SnapshotWindow: StrictStr
    AuthTokenEnabled: StrictBool
    AuthTokenLastModifiedDate: datetime.datetime
    TransitEncryptionEnabled: StrictBool
    AtRestEncryptionEnabled: StrictBool
    ARN: StrictStr
    ReplicationGroupLogDeliveryEnabled: StrictBool
    LogDeliveryConfigurations: list[ElasticacheLogDeliveryConfiguration]
    NetworkType: Literal[
        'ipv4',
        'ipv6',
        'dual_stack',
    ]
    IpDiscovery: Literal[
        'ipv4',
        'ipv6',
    ]
    TransitEncryptionMode: Literal[
        'preferred',
        'required',
    ]


import asyncio
from botocore.response import StreamingBody


class AWSs3StreamingBody:

    def __init__(
        self,
        body: StreamingBody
    ) -> None:
        self.body = body

    async def read(self):
        return await asyncio.to_thread(
            self.body.read
        )
    
    async def readable(self):
        return await asyncio.to_thread(
            self.body.readable
        )
    
    async def readlines(self):
        return await asyncio.to_thread(
            self.body.readline
        )
    
    async def __aiter__(self):
        for chunk in (
            await asyncio.to_thread(
                self.body.__iter__
            )
        ):
            yield chunk

    async def __anext__(self):
        return await asyncio.to_thread(
            self.body.__next__
        )
    
    async def __aexit__(self):
        return await asyncio.to_thread(
            self.body.__aexit__
        )
    
    async def __aenter__(self):
        return await asyncio.to_thread(
            self.body.__enter__
        )
    
    def iter_lines(
        self,
        chunk_size: int=1024,
        keepends: bool=False
    ):
        return asyncio.to_thread(
            self.body.iter_lines,
            chunk_size=chunk_size,
            keepends=keepends
        )
    
    def iter_chunks(
        self,
        chunk_size: int=1024
    ):
        return asyncio.to_thread(
            self.body.iter_chunks,
            chunk_size=chunk_size
        )
    
    async def tell(self):
        return await asyncio.to_thread(
            self.body.tell
        )
    
    async def close(self):
        return await asyncio.to_thread(
            self.body.close
        )


from pydantic import BaseModel, StrictStr
from .elasticache_replication_group import ElasticacheReplicationGroup


class ElasticacheDescribeReplicationGroupsResponse(BaseModel):
    Marker: StrictStr
    ReplicationGroups: list[ElasticacheReplicationGroup]




import datetime
from pydantic import (
    BaseModel,
    StrictStr,
    StrictInt,
    StrictBool,
)

from typing import Literal
from .msk_cluster_broker_az_distribution import MSKClusterBrokerAZDistribution
from .msk_ebs_storage_info import MSKEBSStorageInfo
from .msk_connectivity_info import MSKConnectivityInfo
from .msk_client_authentication import MSKClientAuthentication
from .msk_current_broker_software_info import MSKCurrentBrokerSoftwareInfo
from .msk_encryption_info import MSKEncryptionInfo
from .msk_enhanced_monitoring import MSKEnhancedMonitoring
from .msk_logging_info import MSKLoggingInfo
from .msk_open_monitoring import MSKOpenMonitoring
from .msk_node_group_state import MSKNodeGroupState
from .msk_state_info import MSKStateInfo
from .msk_storage_mode import MSKStorageMode
from .msk_customer_action_status import MSKCustomerActionStatus


class MSKBrokerNodeGroupInfo(BaseModel):
    BrokerAZDistribution: MSKClusterBrokerAZDistribution = 'DEFAULT'
    ClientSubnets: list[StrictStr]
    InstanceType: StrictStr
    SecurityGroups: list[StrictStr]
    StorageInfo: MSKEBSStorageInfo
    ConnectivityInfo: MSKConnectivityInfo



from pydantic import BaseModel
from .msk_client_authentication_sasl import MSKClientAuthenticationSasl
from .msk_client_authentication_tls import MSKClientAuthenticationTls
from .msk_client_authentication_unauthenticated import MSKClientAuthenticationUnauthenticated


class MSKClientAuthentication(BaseModel):
    Sasl: MSKClientAuthenticationSasl
    Tls: MSKClientAuthenticationTls
    Unauthenticated: MSKClientAuthenticationUnauthenticated



from pydantic import BaseModel
from .msk_cloudwatch_logs import MSKCloudWatchLogs
from .msk_firehose import MSKFirehose
from .msk_s3 import MSKS3


class MSKBrokerLogs(BaseModel):
    CloudWatchLogs: MSKCloudWatchLogs
    Firehose: MSKFirehose
    S3: MSKS3


import datetime
from pydantic import BaseModel, StrictStr, StrictInt, StrictBool
from typing import Literal
from .elasticache_endpoint import ElasticacheEndpoint
from .elasticache_node_group_member import ElasticacheNodeGroupMember
from .elasticache_log_delivery_configuration import ElasticacheLogDeliveryConfiguration

ElasticacheAutomaticFailover = Literal[
    "enabled",
    "disabled",
    "enabling",
    "disabling",
]

ElasticacheMultiAZ = Literal[
    "enabled",
    "disabled",
]

ElasticacheDataTiering = Literal[
    "enabled",
    "disabled"
]

ElasticacheNetworkType = Literal[
    "ipv4",
    "ipv6",
    "dual_stack",
]

ElasticacheIpDiscovery = Literal[
    "ipv4",
    "ipv6"
]

ElasticacheTransitEncryptionMode = Literal[
    "preferred",
    "required",
]

ElasticacheClusterMode = Literal[
    "enabled",
    "disabled",
    "compatible"
]


class ElasticacheNodeGroup(BaseModel):
    NodeGroupId: StrictStr
    Status: StrictStr
    PrimaryEndpoint: ElasticacheEndpoint
    ReaderEndpoint: ElasticacheEndpoint
    Slots: StrictStr
    NodeGroupMembers: list[ElasticacheNodeGroupMember] | None = None
    SnapshottingClusterId: StrictStr
    AutomaticFailover: ElasticacheAutomaticFailover
    MultiAZ: ElasticacheMultiAZ
    ConfigurationEndpoint: ElasticacheEndpoint
    SnapshotRetentionLimit: StrictInt
    SnapshotWindow: StrictStr
    ClusterEnabled: StrictBool = True
    CacheNodeType: StrictStr
    AuthTokenEnabled: StrictBool = False
    AuthTokenLastModifiedDate: datetime.datetime
    TransitEncryptionEnabled: StrictBool = False
    AtRestEncryptionEnabled: StrictBool = False
    MemberClusterOutpostArns: list[StrictStr] | None = None
    KmsKeyId: StrictStr
    ARN: StrictStr
    UserGroupIds: list[StrictStr] | None = None
    LogDeliveryConfigurations: list[ElasticacheLogDeliveryConfiguration] | None = None
    ReplicationGroupCreateTime: datetime.datetime
    DataTiering: ElasticacheDataTiering
    AutoMinorVersionUpgrade: StrictBool = False
    NetworkType: ElasticacheNetworkType
    IpDiscovery: ElasticacheIpDiscovery
    TransitEncryptionMode: ElasticacheTransitEncryptionMode
    ClusterMode: ElasticacheClusterMode




from __future__ import annotations
import datetime
from abc import ABC, abstractmethod
from io import FileIO
from typing import (
    Dict, 
    Optional, 
    Literal, 
    Union, 
    List,
    TextIO,
    BinaryIO,
    TypedDict,
    Callable,
    Any
)
from .s3_multipart_upload_part import s3MultipartUploadPart
from .s3_tag import s3Tag


class s3Client(ABC):

    @abstractmethod
    def list_buckets(self) -> Dict[str, str]:
        pass

    @abstractmethod
    def list_objects_v2(
        self,
        Bucket: str=None,
        Delimiter: Optional[str]=None,
        EncodingType: Optional[str]=None,
        MaxKeys: Optional[int]=None,
        Prefix: Optional[str]=None,
        ContinuationToken: Optional[str]=None,
        FetchOwner: Optional[bool]=None,
        StartAfter: Optional[str]=None,
        ExpectedBucketOwner: Optional[str]=None,
        OptionalObjectAttributes: Optional[
            List[str]
        ]=None
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def get_object(
        self,
        Bucket: str=None,
        IfMatch: Optional[str]=None,
        IfModifiedSince: Optional[datetime.datetime]=None,
        IfNoneMatch: Optional[str]=None,
        IfUnmodifiedSince: Optional[datetime.datetime]=None,
        Key: str=None,
        Range: Optional[str]=None,
        ResponseCacheControl: Optional[str]=None,
        ResponseContentDisposition: Optional[str]=None,
        ResponseContentEncoding: Optional[str]=None,
        ResponseContentLanguage: Optional[str]=None,
        ResponseContentType: Optional[str]=None,
        ResponseExpires: Optional[datetime.datetime]=None,
        VersionId: Optional[str]=None,
        SSECustomerAlgorithm: Optional[str]=None,
        SSECustomerKey: Optional[str]=None,
        PartNumber: Optional[int]=None,
        ExpectedBucketOwner: Optional[str]=None,
        ChecksumMode: Union[
            Optional[Literal['ENABLED']],
            Optional[Literal['DISABLED']]
        ]=None
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def put_object(
        self,
        ACL: Union[
            Optional[Literal['private']],
            Optional[Literal['public-read']],
            Optional[Literal['public-read-write']],
            Optional[Literal['authenticated-read']],
            Optional[Literal['aws-exec-read']],
            Optional[Literal['bucket-owner-read']],
            Optional[Literal['bucket-owner-full-control']]
        ]=None,
        Body: Union[bytes, TextIO, BinaryIO]=None,
        Bucket: str=None,
        CacheControl: Optional[str]=None,
        ContentDisposition: Optional[str]=None,
        ContentEncoding: Optional[str]=None,
        ContentLanguage: Optional[str]=None,
        ContentLength: Optional[int]=None,
        ContentMD5: Optional[str]=None,
        ContentType: Optional[str]=None,
        ChecksumAlgorithm: Union[
            Optional[Literal['CRC32']],
            Optional[Literal['CRC32C']],
            Optional[Literal['SHA1']],
            Optional[Literal['SHA256']]
        ]=None,
        ChecksumCRC32: Optional[str]=None,
        ChecksumCRC32C: Optional[str]=None,
        ChecksumSHA1: Optional[str]=None,
        ChecksumSHA256: Optional[str]=None,
        Expires: Optional[datetime.datetime]=None,
        GrantFullControl: Optional[str]=None,
        GrantRead: Optional[str]=None,
        GrantReadACP: Optional[str]=None,
        GrantWriteACP: Optional[str]=None,
        Key: str=None,
        Metadata: Optional[Dict[str, str]]=None,
        ServerSideEncryption: Union[
            Optional[Literal['AES256']],
            Optional[Literal['aws:kms']],
            Optional[Literal['aws:kms:dsse']]
        ]=None,
        StorageClass: Union[
            Optional[Literal['STANDARD']],
            Optional[Literal['REDUCED_REDUNDANCY']],
            Optional[Literal['STANDARD_IA']],
            Optional[Literal['ONEZONE_IA']],
            Optional[Literal['INTELLIGENT_TIERING']],
            Optional[Literal['GLACIER']],
            Optional[Literal['DEEP_ARCHIVE']],
            Optional[Literal['OUTPOSTS']],
            Optional[Literal['GLACIER_IR']],
            Optional[Literal['SNOW']]
        ]=None,
        WebsiteRedirectLocation: Optional[str]=None,
        SSECustomerAlgorithm: Optional[str]=None,
        SSECustomerKey: Optional[str]=None,
        SSEKMSKeyId: Optional[str]=None,
        SSEKMSEncryptionContext: Optional[str]=None,
        BucketKeyEnabled: Optional[bool]=None,
        Tagging: Optional[str]=None,
        ObjectLockMode: Union[
            Optional[Literal['GOVERNANCE']],
            Optional[Literal['COMPLIANCE']]
        ]=None,
        ObjectLockRetainUntilDate: Optional[datetime.datetime]=None,
        ObjectLockLegalHoldStatus: Union[
            Optional[Literal['ON']],
            Optional[Literal['OFF']]
        ]=None,
        ExpectedBucketOwner: Optional[str]=None
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def delete_bucket(
        self,
        Bucket='string',
        ExpectedBucketOwner='string'
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def delete_object(
        self,
        Bucket: str=None,
        Key: str=None,
        MFA: Optional[str]=None,
        VersionId: Optional[str]=None,
        BypassGovernanceRetention: Optional[bool]=None,
        ExpectedBucketOwner: Optional[str]=None
    ) -> Dict[str, str]:
        pass
    
    @abstractmethod
    def list_multipart_uploads(
        self,
        Bucket: str=None,
        Delimiter: Optional[str]=None,
        EncodingType: Optional[str]=None,
        KeyMarker: Optional[str]=None,
        MaxUploads: Optional[int]=None,
        Prefix: Optional[str]=None,
        UploadIdMarker: Optional[str]=None,
        ExpectedBucketOwner: Optional[str]=None
    ) -> Dict[str, str]:
        pass
    
    @abstractmethod
    def list_parts(
        self,
        Bucket: str=None,
        Key: str=None,
        MaxParts: Optional[int]=None,
        PartNumberMarker: Optional[int]=None,
        UploadId: Optional[str]=None,
        ExpectedBucketOwner: Optional[str]=None,
        SSECustomerAlgorithm: Optional[str]=None,
        SSECustomerKey: Optional[str]=None,
    ) -> Dict[str, str]:
        pass
    
    @abstractmethod
    def complete_multipart_upload(
        self,
        Bucket: str=None,
        Key: str=None,
        MultipartUpload: Dict[
             Literal['Parts'],
             List[s3MultipartUploadPart]
        ]=None,
        UploadId: str=None,
        ChecksumCRC32: Optional[str]=None,
        ChecksumCRC32C: Optional[str]=None,
        ChecksumSHA1: Optional[str]=None,
        ChecksumSHA256: Optional[str]=None,
        ExpectedBucketOwner: Optional[str]=None,
        SSECustomerAlgorithm: Optional[str]=None,
        SSECustomerKey: Optional[str]=None,
    ) -> Dict[str, str]:
        pass
    
    @abstractmethod
    def abort_multipart_upload(
        self,
        Bucket: str=None,
        Key: str=None,
        UploadId: str=None,
        ExpectedBucketOwner: Optional[str]=None
    ) -> Dict[str, str]:
        pass
    
    @abstractmethod
    def create_multipart_upload(
        self,
        ACL: Union[
            Optional[Literal['private']],
            Optional[Literal['public-read']],
            Optional[Literal['public-read-write']],
            Optional[Literal['authenticated-read']],
            Optional[Literal['aws-exec-read']],
            Optional[Literal['bucket-owner-read']],
            Optional[Literal['bucket-owner-full-control']]
        ]=None,
        Body: Union[bytes, TextIO, BinaryIO]=None,
        Bucket: str=None,
        CacheControl: Optional[str]=None,
        ContentDisposition: Optional[str]=None,
        ContentEncoding: Optional[str]=None,
        ContentLanguage: Optional[str]=None,
        ContentType: Optional[str]=None,
        Expires: Optional[datetime.datetime]=None,
        GrantFullControl: Optional[str]=None,
        GrantRead: Optional[str]=None,
        GrantReadACP: Optional[str]=None,
        GrantWriteACP: Optional[str]=None,
        Key: str=None,
        Metadata: Optional[Dict[str, str]]=None,
        ServerSideEncryption: Union[
            Optional[Literal['AES256']],
            Optional[Literal['aws:kms']],
            Optional[Literal['aws:kms:dsse']]
        ]=None,
        StorageClass: Union[
            Optional[Literal['STANDARD']],
            Optional[Literal['REDUCED_REDUNDANCY']],
            Optional[Literal['STANDARD_IA']],
            Optional[Literal['ONEZONE_IA']],
            Optional[Literal['INTELLIGENT_TIERING']],
            Optional[Literal['GLACIER']],
            Optional[Literal['DEEP_ARCHIVE']],
            Optional[Literal['OUTPOSTS']],
            Optional[Literal['GLACIER_IR']],
            Optional[Literal['SNOW']]
        ]=None,
        WebsiteRedirectLocation: Optional[str]=None,
        SSECustomerAlgorithm: Optional[str]=None,
        SSECustomerKey: Optional[str]=None,
        SSEKMSKeyId: Optional[str]=None,
        SSEKMSEncryptionContext: Optional[str]=None,
        BucketKeyEnabled: Optional[bool]=None,
        Tagging: Optional[str]=None,
        ObjectLockMode: Union[
            Optional[Literal['GOVERNANCE']],
            Optional[Literal['COMPLIANCE']]
        ]=None,
        ObjectLockRetainUntilDate: Optional[datetime.datetime]=None,
        ObjectLockLegalHoldStatus: Union[
            Optional[Literal['ON']],
            Optional[Literal['OFF']]
        ]=None,
        ExpectedBucketOwner: Optional[str]=None,
        ChecksumAlgorithm: Union[
             Optional[Literal['CRC32']],
             Optional[Literal['CRC32C']],
             Optional[Literal['SHA1']],
             Optional[Literal['SHA256']]
        ]=None
    ) -> Dict[str, str]:
        pass
    
    @abstractmethod
    def upload_part(
        self,
        Body: Union[
             bytes,
             TextIO,
             BinaryIO
        ]=None,
        Bucket: Optional[str]=None,
        ContentLength: Optional[int]=None,
        ContentMD5: Optional[str]=None,
        ContentType: Optional[str]=None,
        ChecksumAlgorithm: Union[
            Optional[Literal['CRC32']],
            Optional[Literal['CRC32C']],
            Optional[Literal['SHA1']],
            Optional[Literal['SHA256']]
        ]=None,
        ChecksumCRC32: Optional[str]=None,
        ChecksumCRC32C: Optional[str]=None,
        ChecksumSHA1: Optional[str]=None,
        ChecksumSHA256: Optional[str]=None,
        Key: str=None,
        PartNumber: int=None,
        UploadId: str=None,
        SSECustomerAlgorithm: Optional[str]=None,
        SSECustomerKey: Optional[str]=None,
        RequestPayer: Optional[
            Literal['requester']
        ]=None,
        ExpectedBucketOwner: Optional[str]=None,
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def upload_part_copy(
        self,
        Bucket: str=None,
        CopySource: str | TypedDict[
            Literal['Bucket','Key','VersionId'],
            str
        ]=None,
        Key: str=None,
        PartNumber: int=None,
        UploadId: str=None,
        SSECustomerAlgorithm: Optional[str]=None,
        SSECustomerKey: Optional[str]=None,
        RequestPayer: Optional[
            Literal['requester']
        ]=None,
        ExpectedBucketOwner: Optional[str]=None
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def get_bucket_tagging(
        self,
        Bucket: str=None,
        ExpectedBucketOwner: Optional[str]=None,
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def get_object_tagging(
        self,
        Bucket: str=None,
        Key: str=None,
        VersionId: Optional[str]=None,
        ExpectedBucketOwner: Optional[str]=None,
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def put_bucket_tagging(
        self,
        Bucket: str=None,
        ChecksumAlgorithm: Union[
            Optional[Literal['CRC32']],
            Optional[Literal['CRC32C']],
            Optional[Literal['SHA1']],
            Optional[Literal['SHA256']]
        ]=None,
        Tagging: Dict[
            Literal['TagSet'],
            List[s3Tag]
        ]=None,
        ExpectedBucketOwner: Optional[str]=None
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def put_object_tagging(
        self,
        Bucket: str=None,
        Key: str=None,
        VersionId: Optional[str]=None,
        ContentMD5: Optional[str]=None,
        ChecksumAlgorithm: Optional[
            Literal[
                'CRC32',
                'CRC32C',
                'SHA1',
                'SHA256'
            ]
        ]=None,
        Tagging: Dict[
            Literal['TagSet'],
            List[s3Tag]
        ]=None,
        ExpectedBucketOwner: Optional[str]=None
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def delete_bucket_tagging(
        self,
        Bucket: str=None,
        ExpectedBucketOwner: Optional[str]=None
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def delete_object_tagging(
        self,
        Bucket: str=None,
        Key: str=None,
        VersionId: Optional[str]=None,
        ExpectedBucketOwner: Optional[str]=None
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def upload_fileobj(
        self,
        Fileobj: FileIO,
        Bucket: str,
        Key: str,
        ExtraArgs: Dict[str, str],
        Callback: Callable[..., Any],
        Config: Dict[str, str]
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def upload_file(
        self,
        Fileobj: FileIO,
        Bucket: str,
        Key: str,
        ExtraArgs: Dict[str, str],
        Callback: Callable[..., Any],
        Config: Dict[str, str]
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def list_directory_buckets(
        self,
        ContinuationToken: str=None,
        MaxDirectoryBuckets: int=None
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def list_object_versions(
        Bucket: str=None,
        Delimiter: str=None,
        EncodingType: Literal['url']=None,
        KeyMarker: str=None,
        MaxKeys: int=None,
        Prefix: str=None,
        VersionIdMarker: str=None,
        ExpectedBucketOwner: str=None,
        RequestPayer: Literal['requester']=None,
        OptionalObjectAttributes: List[str]=None
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def delete_objects(
        self,
        Bucket: str=None,
        Delete: Dict[
            Literal['Objects', 'Quiet'],
            List[
                Dict[
                    Literal['Key', 'VersionId'],
                    str
                ]
            ] | bool
        ]=None,
        MFA: str=None,
        RequestPayer: Literal['requester']=None,
        BypassGovernanceRetention: bool=None,
        ExpectedBucketOwner: str=None,
        ChecksumAlgorithm: Literal[
            'CRC32',
            'CRC32C',
            'SHA1',
            'SHA256'
        ]=None
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def download_file(
        self,
        Bucket: str,
        Key: str,
        Filename: str,
        ExtraArgs: Dict[str, str],
        Callback: Callable[..., Any],
        Config: str
    ) -> None:
        pass

    @abstractmethod
    def download_fileobj(
        self,
        Bucket: str,
        Key: str,
        Fileobj: FileIO,
        ExtraArgs: Dict[str, str],
        Callback: Callable[..., Any],
        Config: str
    ) -> None:
        pass

    @abstractmethod
    def generate_presigned_url(
        self,
        ClientMethod: str,
        Params: Dict[str, str]=None,
        ExpiresIn: int=None,
        HttpMethod: str=None
    )-> str:
        pass

    @abstractmethod
    def generate_presigned_post(
        self,
        Bucket: str,
        Key: str,
        Fields: Dict[str, str]=None,
        Conditions: List[Dict[str, str] | List[str]]=None,
        ExpiresIn: int=None
    ) -> Dict[str, str]:
        pass
    
    @abstractmethod
    def copy_object(
        self,
        ACL: Literal[
            'private',
            'public-read',
            'public-read-write',
            'authenticated-read',
            'aws-exec-read',
            'bucket-owner-read',
            'bucket-owner-full-control'
        ]=None,
        Bucket: str=None,
        CacheControl: str=None,
        ChecksumAlgorithm: Literal[
            'CRC32',
            'CRC32C',
            'SHA1',
            'SHA256'
        ]=None,
        ContentDisposition: str=None,
        ContentEncoding: str=None,
        ContentLanguage: str=None,
        ContentType: str=None,
        CopySource: str | Dict[
            Literal['Bucket', 'Key', 'VersionId'],
            str
        ]=None,
        CopySourceIfMatch: str=None,
        CopySourceIfModifiedSince: datetime.datetime=None,
        CopySourceIfNoneMatch: str=None,
        CopySourceIfUnmodifiedSince: datetime.datetime=None,
        Expires: datetime.datetime=None,
        GrantFullControl: str=None,
        GrantRead: str=None,
        GrantReadACP: str=None,
        GrantWriteACP: str=None,
        Key: str=None,
        Metadata: Dict[str, str]=None,
        MetadataDirective: Literal['COPY', 'REPLACE']=None,
        TaggingDirective: Literal['COPY', 'REPLACE']=None,
        ServerSideEncryption: Literal[
            'AES256',
            'aws:kms',
            'aws:kms:dsse'
        ]=None,
        StorageClass: Literal[
            'STANDARD',
            'REDUCED_REDUNDANCY',
            'STANDARD_IA',
            'ONEZONE_IA',
            'INTELLIGENT_TIERING',
            'GLACIER',
            'DEEP_ARCHIVE',
            'OUTPOSTS',
            'GLACIER_IR',
            'SNOW',
            'EXPRESS_ONEZONE'
        ]=None,
        WebsiteRedirectLocation: str=None,
        SSECustomerAlgorithm: str=None,
        SSECustomerKey: str=None,
        SSEKMSKeyId: str=None,
        SSEKMSEncryptionContext: str=None,
        BucketKeyEnabled: bool=None,
        CopySourceSSECustomerAlgorithm: str=None,
        CopySourceSSECustomerKey: str=None,
        RequestPayer: Literal['requester']=None,
        Tagging: str= None,
        ObjectLockMode: Literal['GOVERNANCE', 'COMPLIANCE']=None,
        ObjectLockRetainUntilDate: datetime.datetime=None,
        ObjectLockLegalHoldStatus: Literal['ON', 'OFF']=None,
        ExpectedBucketOwner: str=None,
        ExpectedSourceBucketOwner: str=None
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def copy(
        self,
        copy_source: Dict[
            Literal[
                'Bucket',
                'Key',
                'VersionId'
            ],
            str
        ],
        Bucket: str,
        Key: str,
        ExtraArgs: Dict[str, str]=None,
        Callback: Callable[..., Any]=None,
        SourceClient: s3Client=None,
        Config: Dict[str, str]=None
    ) -> None:
        pass

    @abstractmethod
    def put_bucket_versioning(
        self,
        Bucket: str=None,
        ChecksumAlgorithm: Literal[
            'CRC32',
            'CRC32C',
            'SHA1',
            'SHA256'
        ]=None,
        MFA: str=None,
        VersioningConfiguration: Dict[
            Literal['MFADelete', 'Status'],
            Literal['Enabled', 'Disabled'] |
            Literal['Enabled', 'Suspended']
        ]=None,
        ExpectedBucketOwner: str=None
    ) -> Dict[str, str]:
        pass

    @abstractmethod
    def close(self):
        pass


from pydantic import (
    BaseModel,
    StrictInt,
    StrictStr,
)


class OrganizationsListAccountsRequest(BaseModel):
    NextToken: StrictStr | None = None
    MaxResults: StrictInt | None = None


from pydantic import BaseModel, StrictStr
from .ec2_volume import EC2Volume


class EC2DescribeVolumesResponse(BaseModel):
    NextToken: StrictStr | None = None
    Volumes: list[EC2Volume] | None = None


from pydantic import BaseModel
from .msk_encryption_at_rest import MSKEncryptionAtRest
from .msk_encryption_in_transit import MSKEncryptionInTransit


class MSKEncryptionInfo(BaseModel):
    EncryptionAtRest: MSKEncryptionAtRest
    EncryptionInTransit: MSKEncryptionInTransit


from .msk_list_clusters_request import MSKListClustersRequest as MSKListClustersRequest
from .msk_list_clusters_response import MSKListClustersResponse as MSKListClustersResponse
from .msk_cluster import MSKCluster as MSKCluster


from pydantic import BaseModel, StrictStr, StrictInt, StrictBool


class ElasticacheDescribeCacheClustersRequest(BaseModel):
        CacheClusterId: StrictStr = None
        MaxRecords: StrictInt = 100
        Marker: StrictStr | None = None
        ShowCacheNodeInfo: StrictBool = True
        ShowCacheClustersNotInReplicationGroups: StrictBool = True
